
#define NAME "ju: copyleft 1991 Taiji Yamada\n"
#define THIS "kanji code checker for dos and unix.\n"
#define HELP "-l: check each line, -t: show the code table of inputs.\n"

#include <stdio.h>
#include <jstring.h>
#include <jctype.h>
#include <option.h>
#include <files.h>
#include <escs.h>
#include <jutil.h>

void init_code_table( void )
{
  int n, x, y;

  cls();
  for ( n=0; n<3; n++ ){
    locate( 3 -1 +20*n, 4 -2 );
    printf( "%s CODE:", (n==0)?"ANK":(n==1)?"UPP":"LOW" );
    for ( x=0; x<=0xf; x++ ){
      locate( 3+ x +20*n, 4 -1 ); printf( "%1X", x );
    }
    for ( y=0; y<=0xf; y++ ){
      locate( 3 -1 +20*n, 4+ y ); printf( "%1X", y );
    }
  }
}
void close_code_table()
{
  locate( 0, 20 );
}
void put_code_table( unsigned char c )
{
  locate( 3+ (int)(c>>4), 4+ (int)(c&0x0f) );
  if ( isprint( c ) || iskana( c ) )
    printf( "%c", c );
  else
    printf( "*" );
}
void put_jcode_table( unsigned short c )
{
  unsigned char c1, c2;

  c1 = codec1(c);
  c2 = codec2(c);
  locate( 3+ (int)(c1>>4) +20*1, 4+ (int)(c1&0x0f) ); printf( "*" );
  locate( 3+ (int)(c2>>4) +20*2, 4+ (int)(c2&0x0f) ); printf( "*" );
}

void ank_table( unsigned char *src )
{
  while ( *src != '\0' ){
    put_code_table( *src++ );
  }
}

void sj_table( unsigned char *src )
{
  unsigned short c;

  locate( 62, 2 ); printf( "SJ :" );
  while ( *src != '\0' ){
    if ( !iskana(*src) && !(iskanji(*src)&&iskanji2(*(src+1))) ){
      put_code_table( *src++ );
    }else{
      if ( iskana(*src) ){
        put_code_table(*src++);
        continue;
      }
      c = s2code(src);
      locate( 66, 2 );
      printf( "%c%c", codec1(c), codec2(c) );
      put_jcode_table( c );
      src += 2;
    }
  }
}

void jis_table( unsigned char *src )
{
  unsigned short c;
  int code;

  locate( 62, 2 ); printf( "JIS:" );
  code = JCT_ANK;
  if ( strstr( src, _JISB ) > strstr( src, _JISE ) )
    code = JCT_JIS;
  while ( *src != '\0' ){
    if ( isjise(src) ){
      code = JCT_ANK;
      src += 3;
    }else if ( isjiss(src) ){
      code = JCT_JIS;
      src += 3;
    }else{
      if ( code == JCT_ANK ){
        put_code_table( *src++ );
      }else{
        c = s2code(src);
        locate( 66, 2 );
        printf( "%s%c%c%s", _JISB, codec1(c), codec2(c), _JISE );
        put_jcode_table( c );
        src += 2;
      }
    }
  }
}

void euc_table( unsigned char *src )
{
  unsigned short c;

  locate( 62, 2 ); printf( "EUC:" );
  while ( *src != '\0' ){
    if ( !(iseuc(*src)&&iseuc2(*(src+1))) ){
      put_code_table( *src++ );
    }else{
      c = s2code(src);
      locate( 66, 2 );
      printf( "%c%c", codec1(c), codec2(c) );
      put_jcode_table( c );
      src += 2;
    }
  }
}

void main( int argc, char *argv[] )
{
  enum modes { CODE_CHECK, LINE_CHECK, CODE_TABLE };
  int code, mode = CODE_CHECK;
  char buf[2048];
  char *fn;
  FILE *fp;

  if ( scan_option(argc,argv,"-l") ) mode = LINE_CHECK;
  if ( scan_option(argc,argv,"-t") ) mode = CODE_TABLE;
  if ( scan_option(argc,argv,"-help") || scan_option(argc,argv,"/?") ){
    fprintf( stderr, NAME );
    fprintf( stderr, THIS );
    fprintf( stderr, HELP );
    exit( 0 );
  }
  if ( (fn=read_file(argc,argv)) != NULL ){
    /* In the case of file reading, the kanji code is checked exactly. */
    fp = fopen( fn, "r" );
    code = jfcode( fp );
    switch ( mode ){
    case CODE_CHECK:
      switch ( code ){
      case JCT_ANK: printf( "ank.\n" ); break;
      case JCT_SJ:  printf( "sj. \n" ); break;
      case JCT_JIS: printf( "jis.\n" ); break;
      case JCT_EUC: printf( "euc.\n" ); break;
      }
      break;
    case LINE_CHECK:
      switch ( code ){
      case JCT_ANK: printf( "ank.\n" ); break;
      case JCT_SJ:  printf( "sj. \n" ); break;
      case JCT_JIS: printf( "jis.\n" ); break;
      case JCT_EUC: printf( "euc.\n" ); break;
      }
      while ( fgets(buf,2048,fp) != NULL ){
        code = jcode( buf );
        switch ( code ){
        case JCT_ANK: printf( "ANK" ); break;
        case JCT_SJ:  printf( "SJ " ); break;
        case JCT_JIS: printf( "JIS" ); break;
        case JCT_EUC: printf( "EUC" ); break;
        }
        printf( ":%s", buf );
      }
      break;
    case CODE_TABLE:
      init_code_table();
      while ( fgets(buf,2048,fp) != NULL ){
        switch ( code ){
        case JCT_ANK: ank_table( buf ); break;
        case JCT_SJ:  sj_table ( buf ); break;
        case JCT_JIS: jis_table( buf ); break;
        case JCT_EUC: euc_table( buf ); break;
        }
      }
      close_code_table();
      break;
    }
  }else{
    /* In the case of filter, the kanji code is checked not exactly. */
    fp = stdin;
    switch ( mode ){
    case CODE_CHECK:
      code = jfcode( fp );
      switch ( code ){
      case JCT_ANK: printf( "ank.\n" ); break;
      case JCT_SJ:  printf( "sj. \n" ); break;
      case JCT_JIS: printf( "jis.\n" ); break;
      case JCT_EUC: printf( "euc.\n" ); break;
      }
      break;
    case LINE_CHECK:
      while ( fgets(buf,2048,fp) != NULL ){
        code = jcode( buf );
        switch ( code ){
        case JCT_ANK: printf( "ANK" ); break;
        case JCT_SJ:  printf( "SJ " ); break;
        case JCT_JIS: printf( "JIS" ); break;
        case JCT_EUC: printf( "EUC" ); break;
        }
        printf( ":%s", buf );
      }
      break;
    case CODE_TABLE:
      init_code_table();
      while ( fgets(buf,2048,fp) != NULL ){
        code = jcode( buf );
        switch ( code ){
        case JCT_ANK: ank_table( buf ); break;
        case JCT_SJ:  sj_table ( buf ); break;
        case JCT_JIS: jis_table( buf ); break;
        case JCT_EUC: euc_table( buf ); break;
        }
      }
      close_code_table();
      break;
    }
  }
  exit( 0 );
}
