/*
 * routines to analyse command line options < option.h >
 *
 * ޥɥ饤󥪥ץϤΤΥ롼
 *
 * Copyright (c) 1992-94 Tokyo Denki University, Taiji Yamada
 */
#ifndef _OPTION_H_
#define _OPTION_H_
#include <stdarg.h>
#include <string.h>
#include <malloc.h>

/*
 * 1. ޥɥ饤󥪥ץ˻ꤵ줿 key ¸ߤ, ֤.
 * () ޥɥ饤 "-help" ꤵƤ
 *      scan_option( argc, argv, "-help" ) ֤, .
 *
 * 2. ޥɥ饤󥪥ץ˻ꤵ줿 key ¸ߤ,
 *   key  scanf 񼰤ľ, key ΤȤ˻ꤵ줿Ѱ
 *  ޥɥ饤ͤ򥻥åȤ, ֤.
 * () ޥɥ饤 "-geom 100x100" ˤ
 *       scan_option( argc, argv, "-geom %dx%d", &W, &H ) ֤,
 *       W, H ˤϤ줾 100, 100 .
 */
/*
 * 1. if there is key to scan in command line options, true is returned.
 * (ex.) if there is "-help" in command line options
 *       scan_option( argc, argv, "-help" ) returns !0
 *
 * 2. if there is key to scan with scanf format,
 *   value(s) is(are) scaned and !0 is returned.
 * (ex.) if there is "-geom 100x100" in command line options
 *       scan_option( argc, argv, "-geom %dx%d", &W, &H ) returns !0 and
 *       sets W, H to 100, 100 respectively.
 */
static int scan_option( int argc, char *argv[], char *key, ... )
{
  int i, idx;
  char opt[64], fmt[64], buf[256], *ptr;
  va_list argptr;
#ifndef ENABLE_VSSCANF
#define VLIST10(V) V[0],V[1],V[2],V[3],V[4],V[5],V[6],V[7],V[8],V[9]
  int c;
  void **val;
#endif

  strcpy( opt, key );
  if ( (ptr=strchr(opt,'%')) != NULL ){
    strcpy( fmt, ptr );
    *ptr = '\0';
    while ( *(--ptr)==' ' ) *ptr = '\0';
  }else{
    for ( i=1; i<argc; i++ )
      if ( strcmp( argv[i], key )==0 )
        return i;
    return( 0 );
  }

#ifndef ENABLE_VSSCANF
  c = 0;
  ptr = strchr( fmt, '%' );
  while ( ptr != NULL ){
    c++;
    ptr = strchr( ptr+1, '%' );
  }
  val = (void**)malloc( (c+1)*sizeof(void*) );
  va_start( argptr, key );
  for ( i=0; i<c; i++ ){
    val[i] = va_arg( argptr, void* );
  }
  val[i] = NULL;
  va_end( argptr );
#endif

  idx = 0;
  for ( i=1; i<argc; i++ )
    if ( strcmp( argv[i], opt )==0 ){
      strcpy( buf, "" );
      idx = i;
      while ( ++i < argc ){
        strcat( buf, argv[i] );
        strcat( buf, " " );
      }
#ifdef ENABLE_VSSCANF
      va_start( argptr, key );
      vsscanf( buf, fmt, argptr );
      va_end( argptr );
#else
      sscanf( buf, fmt, VLIST10(val) );
#endif
      break;
    }
#ifndef ENABLE_VSSCANF
  free( val );
#undef VLIST10
#endif
  return idx;
  /*
   * following methods do not work in place of vsscanf
   *  sscanf ( buf, fmt, _va_ptr );
   *  sscanf ( buf, fmt, argptr  );
   *  vfscanf( buf, fmt, argptr  );
   */
}

/*
 * scan_option ȤӤΰۤʤ륪ץϥ롼
 * 1. set_option( key, ... )     key ⤷ key  format Ͽ롼.
 * 2. scan_options( argc, argv ) ޥɥ饤ϥ롼.
 * 3. is_option( key )           ޥɥ饤Ϸ̼롼
 * 4. reset_options()            ΰ襯ꥢ롼
 * ()
 * set_option( "-help" );
 * set_option( "-geom %dx%d", &W, &H );
 * scan_options( argc, argv ); ⤷ץ "-geom 100x100" ,
 *                            W, H ˤ 100, 100 .
 * is_option( "-help" ); ץ "-help"  !0 ֤.
 * is_option( "-geom" ); ץ "-geom"  !0 ֤.
 * reset_options(); 
 */
/*
 * The set of analizing routines another scan_option.
 * 1. set_option( key, ... )     is setting routine of key, key and format.
 * 2. scan_options( argc, argv ) is analizing routine of command line options.
 * 3. is_option( key )           is returning routine of results on the key.
 * 4. reset_options()            is cleaning routine of internal memory.
 * (ex.)
 * set_option( "-help" );
 * set_option( "-geom %dx%d", &W, &H );
 * scan_options( argc, argv ); if there is option "-geom 100x100", then W and H
 *                            are setted the values 100 and 100, respectively.
 * is_option( "-help" ); this returns !0 as there is option "-help".
 * is_option( "-geom" ); this returns !0 as there is option "-geom".
 * reset_options(); 
 */
typedef struct _option_t {
  char  *key;
  char  *fmt;
  int    c;
  void **val;
  int    idx;
} option_t;
static int     _n_options = 0;
static option_t *_options = NULL;

static option_t *set_option( char *key, ... )
{
  int i, n;
  char opt[64], *ptr;
  option_t *it;
  va_list argptr;

  n = _n_options;
  _n_options++;
  _options = (option_t*)realloc( _options, _n_options*sizeof(option_t) );
  if ( _options == NULL ) return NULL;
  it = &_options[n];

  strcpy( opt, key );
  if ( (ptr=strchr(opt,'%')) != NULL ){
    it->fmt = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
    strcpy( it->fmt, ptr );
    *ptr = '\0';
    while ( *(--ptr)==' ' ) *ptr = '\0';
  }else{
    it->fmt = NULL;
  }
  it->key = (char*)malloc( (strlen(opt)+1)*sizeof(char) );
  strcpy( it->key, opt );

  it->c = 0;
  it->val = NULL;
  it->idx = 0;
  if ( it->fmt != NULL ){
    ptr = strchr( it->fmt, '%' );
    while ( ptr != NULL ){
      it->c++;
      ptr = strchr( ptr+1, '%' );
    }
    it->val = (void**)malloc( (it->c+1)*sizeof(void*) );
    va_start( argptr, key );
    for ( i=0; i<it->c; i++ ){
      it->val[i] = va_arg( argptr, void* );
    }
    it->val[i] = NULL;
    va_end( argptr );
  }
  return it;
}
static void scan_options( int argc, char *argv[] )
{
#define VLIST10(V) V[0],V[1],V[2],V[3],V[4],V[5],V[6],V[7],V[8],V[9]
  int i, n;
  char buf[256];
  option_t *it;

  for ( n=0; n<_n_options; n++ ){
    it = &_options[n];
    for ( i=1; i<argc; i++ ){
      if ( strcmp( argv[i], it->key )==0 ){
        it->idx = i;
        break;
      }
    }
    if ( it->fmt != NULL && it->idx != 0 ){
      strcpy( buf, "" );
      i = it->idx;
      while ( ++i < argc ){
        strcat( buf, argv[i] );
        strcat( buf, " " );
      }
      sscanf( buf, it->fmt, VLIST10(it->val) );
    }
  }
#undef VLIST10
}

static int is_option( char *key )
{
  int i, n;
  option_t *it;

  for ( n=0; n<_n_options; n++ ){
    it = &_options[n];
    if ( strcmp( key, it->key ) == 0 ){
      return it->idx;
    }
  }
  return 0;
}

static void reset_options( void )
{
  int i, n;
  option_t *it;

  for ( n=0; n<_n_options; n++ ){
    it = &_options[n];
    if ( it->c != 0 )
      free( it->val );
    if ( it->fmt != NULL ) free( it->fmt );
    if ( it->key != NULL ) free( it->key );
  }
  free( _options );
  _options = NULL;
  _n_options = 0;
}

#endif /* _OPTION_H_ */
