/*
 * Source File < graphics.c >
 *
 * Standard Functions: Borland Graphics Interface for Xlib
 *
 * Copyright (c) 1992-94 Tokyo Denki University, Taiji Yamada
 */
#include <malloc.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "option.h"
#include "graphix.h"
#include "graphics.h"
#include "xbgiutl.h"
#include "bgiutl.h"
#include "_bgi.h"

/*
 * the header of XBGI to stack and current XBGI for graphics
 */
xbgi *XBGIS = NULL;
xbgi *XBGI  = NULL;

/*
 * the registered XBGI driver
 */
void (*XBGI_DRIVER)(void) = XBGI_driver;

/*
 * official enhanced functions for X interface
 */
void xflush( void )
{
  XFlush( XBGI->dpy );
}
void settitle( char *title )
{
  XStoreName( XBGI->dpy, XBGI->root, title );
  XSetIconName( XBGI->dpy, XBGI->root, title );
}
void seticonname( char *filename )
{
  XSetIconName( XBGI->dpy, XBGI->root, filename );
}
static int XSetIconFromPixmap( Display *d, Window r,
                              Pixmap pm, unsigned int w, unsigned int h )
{
  int i, j, n;
  XIconSize *is;
  int status;
  Bool wok = False, hok = False;
  XWMHints wmh;
  
  status = XGetIconSizes( d, r, &is, &n );
  if ( status == 0 ) return 0;
  for ( i=0; i<n; i++ ){
    for ( j=is[i].min_width; j<=is[i].max_width; j+=is[i].width_inc )
      if ( w == j ){
        wok = True;
        break;
      }
    for ( j=is[i].min_height; j<=is[i].max_height; j+=is[i].height_inc )
      if ( h == j ){
        hok = True;
        break;
      }
  }
  if ( wok && hok ){
    wmh.flags = IconPixmapHint;
    wmh.icon_pixmap = pm;
    XSetWMHints( d, r, &wmh );
    return !0;
  }else{
    return 0;
  }
}
int seticon( char *filename )
{
  Pixmap bm;
  unsigned int w, h;
  int xh, yh, status;
  
  status = XReadBitmapFile( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
                           filename,
                           &w, &h, &bm, &xh, &yh );
  if ( status != BitmapSuccess ) return 0;
  status = XSetIconFromPixmap( XBGI->dpy, XBGI->root, bm, w, h );
  XFreePixmap( XBGI->dpy, bm );
  return status;
}
void setmaxx( int maxx )
{
  Window root;
  int i, x, y, width, height, border, depth;
  
  if ( maxx <= 0 ) return;
  if ( XBGI->apage == 0 ){
    XGetGeometry( XBGI->dpy, XBGI->win[0], &root, /* win[0] or root */
                 &x, &y, &width, &height, &border, &depth );
    XResizeWindow( XBGI->dpy, XBGI->root, maxx+1, height );
    for ( i=0; i<XBGI->npage; i++ )
      XResizeWindow( XBGI->dpy, XBGI->win[i], maxx+1, height );
  }else{
    XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
                 &x, &y, &width, &height, &border, &depth );
    XResizeWindow( XBGI->dpy, XBGI->win[XBGI->apage], maxx+1, height );
  }
}
void setmaxy( int maxy )
{
  Window root;
  int i, x, y, width, height, border, depth;

  if ( maxy <= 0 ) return;
  if ( XBGI->apage == 0 ){
    XGetGeometry( XBGI->dpy, XBGI->win[0], &root, /* win[0] or root */
                 &x, &y, &width, &height, &border, &depth );
    XResizeWindow( XBGI->dpy, XBGI->root, width, maxy+1 );
    for ( i=0; i<XBGI->npage; i++ )
      XResizeWindow( XBGI->dpy, XBGI->win[i], width, maxy+1 );
  }else{
    XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
                 &x, &y, &width, &height, &border, &depth );
    XResizeWindow( XBGI->dpy, XBGI->win[XBGI->apage], width, maxy+1 );
  }
}
void getsizexy( int *xsize, int *ysize )
{
  Window root;
  int i, x, y, width, height, border, depth;
  
  if ( XBGI->apage == 0 ){
    XGetGeometry( XBGI->dpy, XBGI->root, &root,
                 &x, &y, &width, &height, &border, &depth );
  }else{
    XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
                 &x, &y, &width, &height, &border, &depth );
  }
  *xsize = width;
  *ysize = height;
}
void setsizexy( int xsize, int ysize )
{
  int i;
  
  if ( xsize <= 0 || ysize <= 0 ) return;
  if ( XBGI->apage == 0 ){
    for ( i=0; i<XBGI->npage; i++ )
      XResizeWindow( XBGI->dpy, XBGI->win[i], xsize, ysize );
    XResizeWindow( XBGI->dpy, XBGI->root, xsize, ysize );
  }else{
    XResizeWindow( XBGI->dpy, XBGI->win[XBGI->apage], xsize, ysize );
  }
}
void getgeometry( int *x, int *y, int *width, int *height )
{
  Window root;
  int border, depth;
  
  XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
               x, y, width, height, &border, &depth );
}
void setgeometry( int x, int y, int width, int height )
{
  XMoveWindow( XBGI->dpy, XBGI->win[XBGI->apage], x, y );
  XResizeWindow( XBGI->dpy, XBGI->win[XBGI->apage], width, height );
}
void setmaxcolor( int c )
{
  int i, cc;
  unsigned long plane_mask;
  struct rgbtype rgb[8] = {
    { 0, 0, 0 },
    { 0, 0, 1 },
    { 0, 1, 0 },
    { 0, 1, 1 },
    { 1, 0, 0 },
    { 1, 0, 1 },
    { 1, 1, 0 },
    { 1, 1, 1 },
  };
  Status ok;
  
  switch( XBGI->gmode ){
  case BGIMONO:
    XBGI->palette.size = 2;
    XBGI->colortable[BlackPixel(XBGI->dpy,XBGI->scr)] = BLACK;
    XBGI->colortable[WhitePixel(XBGI->dpy,XBGI->scr)] = WHITE;
    XBGI->pixels[BLACK] = BlackPixel( XBGI->dpy, XBGI->scr );
    XBGI->pixels[WHITE] = WhitePixel( XBGI->dpy, XBGI->scr );
    for ( i=0; i<XBGI->palette.size; i++ )
      XBGI->palette.colors[i] = i;
    break;
  case BGIGRAY:
    /* I don't know grayscale visual class */
    XBGI->palette.size = c + 1;
    XAllocColorCells( XBGI->dpy, XBGI->colormap,
                     False, &plane_mask, 0, XBGI->pixels, XBGI->palette.size );
    for ( i=0; i<XBGI->palette.size; i++ ){
      XBGI->colortable[XBGI->pixels[i]] = i;
      XBGI->palette.colors[i] = i;
    }
    for ( i=1; i<XBGI->palette.size; i++ )
      setrgbpalette( i, i, i, i );
    setallpalette( &XBGI->palette );
    break;
  case BGICOLOR:
    if ( c+1<1 || 256<c+1 ){
      return;
    }
    if ( XBGI->gerror != grNoInitGraph )
      XFreeColors( XBGI->dpy, XBGI->colormap, XBGI->pixels,
                  XBGI->palette.size, 0 );
    XBGI->palette.size = c + 1;
    ok = XAllocColorCells( XBGI->dpy, XBGI->colormap,
                          False, &plane_mask, False,
                          XBGI->pixels, XBGI->palette.size );
    if ( ok == False ){
      /* 1st failed */
      /* in my case, as size = 232, it is failed. */
      XBGI->colormap = XCreateColormap( XBGI->dpy, XBGI->root, XBGI->vis,
                                       AllocNone );
      /*XInstallColormap( XBGI->dpy, XBGI->colormap );*/ 
      XSetWindowColormap( XBGI->dpy, XBGI->root, XBGI->colormap );
      ok = XAllocColorCells( XBGI->dpy, XBGI->colormap,
                            False, &plane_mask, False,
                            XBGI->pixels, XBGI->palette.size );
      /* technique for settings near the standard color map */
      if ( ok == False ){
        /* 2nd failed */
        /* do nothing */
      }
    }
    for ( i=0; i<XBGI->palette.size; i++ ){
      XBGI->colortable[XBGI->pixels[i]] = i;
      XBGI->palette.colors[i] = i;
    }
    for ( i=0; i<XBGI->palette.size/8; i++ )
      for ( cc=0; cc<8; cc++ )
        setrgbpalette( 8*i+cc,
                      8*(i+1)*rgb[cc].r,
                      8*(i+1)*rgb[cc].g,
                      8*(i+1)*rgb[cc].b );
    for ( i=1; i<XBGI->palette.size/8; i++ )
      setrgbpalette( 8*i+0, 4*(i+1), 4*(i+1), 4*(i+1) );
    cc = XBGI->palette.size/2;
    for ( i=0; i<XBGI->palette.size/8; i++ )
      setrgbpalette( 8*i+7, cc+4*(i+1), cc+4*(i+1), cc+4*(i+1) );
    break;
  }
}
void setbackingstore( int flg )
{
  int i;
  XSetWindowAttributes att;

  if ( flg )
    att.backing_store = Always;
  else
    att.backing_store = NotUseful;
  XChangeWindowAttributes( XBGI->dpy, XBGI->win[XBGI->apage],
                          CWBackingStore, &att );
}
int graphnumber( void )
{
  return( XBGI->id );
}
void selectgraph( int id )
{
  xbgi *ptr = XBGIS;

  while ( ptr != NULL ){
    if ( ptr->id == id ){
      XBGI = ptr;
      return;
    }
    ptr = ptr->next;
  }
}
void termgraph( int id )
{
  xbgi *ptr = XBGIS;

  while ( ptr != NULL ){
    if ( ptr->id == id ){
      destroygraph( ptr );
      return;
    }
    ptr = ptr->next;
  }
}

/*
 *  functions for draw line
 */
int getx( void )
{
  return( XBGI->cp.x );
}
int gety( void )
{
  return( XBGI->cp.y );
}
void getlinesettings( struct linesettingstype *lineinfo )
{
  *lineinfo = XBGI->lineset;
}
/*static void showbit( unsigned pattern )
{
  int i;

  for ( i=0; i<16; i++ )
    if ( pattern&(0x8000>>i) )
      printf( "1" );
    else
      printf( "0" );
}*/
static void upattern2xlinestyle( unsigned upattern,
                                int *offset, char *usrbit, int *n )
{
  int i, j, c, ns, flg;
  unsigned long ul;

  /*showbit( upattern );
  printf( "\n" );*/
  ul = (unsigned long)upattern*0x10000 + upattern;
  ns = 0;
  while ( (!(ul&(0x80000000>>ns)))&&(ns<32) ) ns++;
  if ( ns == 32 ){
    *n = 0;
    return;
  }
  upattern = (ul<<ns)>>16;
  /*showbit( upattern );
  printf( "\n" );*/
  *offset = i = c = 0;
  flg = !0;
  while ( i < 16-ns ){
    if ( flg ){
      if ( !(upattern&(0x8000>>i)) ){
        c++;
        flg = !flg;
      }
      usrbit[c]++;
    }else{
      if ( upattern&(0x8000>>i) ){
        c++;
        flg = !flg;
      }
      usrbit[c]++;
    }
    i++;
  }
  c++;
  if ( ns > 0 ){
    if ( !flg ){
      usrbit[c] = 1; /* ? */
      c++;
    }
    usrbit[c] = ns;
    c++;
    *offset = c-1;
  }
  *n = c;
  /*printf( "offset(%d)", *offset );
  for ( i=0; i<c; i++ )
    printf( " %u", usrbit[i] );
  printf( "num(%d)\n", *n );*/
}
void setlinestyle( int linestyle, unsigned upattern, int thickness )
{
  static char dotted[] = { 1, 3 };
  static char center[] = { 1, 3, 4, 3 };
  static char dashed[] = { 7, 4 };
  char usrbit[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  int offset = 0, n = 0;

  XBGI->lineset.linestyle = linestyle;
  XBGI->lineset.upattern  = upattern;
  XBGI->lineset.thickness = thickness;
  switch( linestyle ){
  case SOLID_LINE:
    break;
  case DOTTED_LINE:
    XSetDashes( XBGI->dpy, XBGI->gcdraw, 0, dotted, 2 );
    break;
  case CENTER_LINE:
    XSetDashes( XBGI->dpy, XBGI->gcdraw, 0, center, 4 );
    break;
  case DASHED_LINE:
    XSetDashes( XBGI->dpy, XBGI->gcdraw, 0, dashed, 2 );
    break;
  case USERBIT_LINE:
    upattern2xlinestyle( upattern, &offset, usrbit, &n );
    if ( n != 0 )
      XSetDashes( XBGI->dpy, XBGI->gcdraw, offset, usrbit, n );
    else
      NotSupported( "setlinestyle( USERBIT_LINE, 0, )" );
    break;
  }
  XSetLineAttributes( XBGI->dpy, XBGI->gcdraw, thickness,
                     ((linestyle==SOLID_LINE)?LineSolid:LineOnOffDash),
                     CapButt, JoinBevel );
}
void line( int x1, int y1, int x2, int y2 )
{
  XDrawLine( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
            Xx( x1 ), Xy( y1 ), Xx( x2 ), Xy( y2 ) );
}
void linerel( int dx, int dy )
{
  XDrawLine( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
            Xx( XBGI->cp.x ), Xy( XBGI->cp.y ),
            Xx( XBGI->cp.x + dx ), Xy( XBGI->cp.y + dy ) );
  XBGI->cp.x += dx;
  XBGI->cp.y += dy;
}
void lineto( int x, int y )
{
  XDrawLine( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
            Xx( XBGI->cp.x ), Xy( XBGI->cp.y ), Xx( x ), Xy( y ) );
  XBGI->cp.x = x;
  XBGI->cp.y = y;
}
void moverel( int dx, int dy )
{
  XBGI->cp.x += dx;
  XBGI->cp.y += dy;
}
void moveto( int x, int y )
{
  XBGI->cp.x = x;
  XBGI->cp.y = y;
}

/*
 * functions for fillstyle
 */
#include "bitmaps/fill/empty.xbm"
#include "bitmaps/fill/solid.xbm"
#include "bitmaps/fill/line.xbm"
#include "bitmaps/fill/ltslsh.xbm"
#include "bitmaps/fill/slsh.xbm"
#include "bitmaps/fill/bkslsh.xbm"
#include "bitmaps/fill/ltbkslsh.xbm"
#include "bitmaps/fill/hatch.xbm"
#include "bitmaps/fill/xhatch.xbm"
#include "bitmaps/fill/intleave.xbm"
#include "bitmaps/fill/widedot.xbm"
#include "bitmaps/fill/closedot.xbm"
#include "bitmaps/fill/user.xbm"
static char *fill_bits[USER_FILL+1] = {
  empty_bits,
  solid_bits,
  line_bits,
  ltslsh_bits,
  slsh_bits,
  bkslsh_bits,
  ltbkslsh_bits,
  hatch_bits,
  xhatch_bits,
  intleave_bits,
  widedot_bits,
  closedot_bits,
  user_bits,
};
void getfillpattern( char *pattern )
{
  int i;

  for ( i=0; i<8; i++ )
    pattern[i] = fill_bits[XBGI->fillset.pattern][i];
}
void getfillsettings( struct fillsettingstype *fillinfo )
{
  *fillinfo = XBGI->fillset;
}
void setfillpattern( char *upattern, int color )
{
  int i;
  Pixmap pm_fill;

  XBGI->fillset.pattern = USER_FILL;
  XBGI->fillset.color   = minmax( 0, color, XBGI->palette.size-1 );
  for ( i=0; i<8; i++ )
    user_bits[i] = upattern[i];
  XSetForeground( XBGI->dpy, XBGI->gcfill, XBGI->pixels[XBGI->fillset.color] );
  pm_fill = XCreateBitmapFromData( XBGI->dpy, XBGI->root,
                                  user_bits, 8, 8 );
  XSetStipple( XBGI->dpy, XBGI->gcfill, pm_fill );
  XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillOpaqueStippled );
}
void setfillstyle( int pattern, int color )
{
  Pixmap pm_fill;

  XBGI->fillset.pattern = minmax( 0, pattern, USER_FILL );
  XBGI->fillset.color   = minmax( 0, color, XBGI->palette.size-1 );
  XSetForeground( XBGI->dpy, XBGI->gcfill, XBGI->pixels[XBGI->fillset.color] );
  switch( XBGI->fillset.pattern ){
  case EMPTY_FILL:
    pm_fill = XCreateBitmapFromData( XBGI->dpy, XBGI->root,
                                    fill_bits[XBGI->fillset.pattern], 8, 8 );
    XSetStipple( XBGI->dpy, XBGI->gcfill, pm_fill );
    XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillStippled );
    break;
  case SOLID_FILL:
    XSetForeground( XBGI->dpy, XBGI->gcfill,
                   XBGI->pixels[XBGI->fillset.color] );
    XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillSolid );
    break;
  case LINE_FILL:
  case LTSLASH_FILL:
  case SLASH_FILL:
  case BKSLASH_FILL:
  case LTBKSLASH_FILL:
  case HATCH_FILL:
  case XHATCH_FILL:
  case INTERLEAVE_FILL:
  case WIDE_DOT_FILL:
  case CLOSE_DOT_FILL:
  case USER_FILL:
    pm_fill = XCreateBitmapFromData( XBGI->dpy, XBGI->root,
                                    fill_bits[XBGI->fillset.pattern], 8, 8 );
    XSetStipple( XBGI->dpy, XBGI->gcfill, pm_fill );
    XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillOpaqueStippled );
    break;
  }
}
void floodfill( int x, int y, int border )
{
#define FLOODFILL_FAST_VERSION
#ifndef FLOODFILL_FAST_VERSION
  int s[1000][2];
  int p, cx, cy, u, u2, d, d2, pu, pd;
  int minx, miny, maxx, maxy;
  int x, y, width, height;

  getgeometry( &x, &y, &width, &height );
  if ( XBGI->viewport.clip ){
    minx = XBGI->viewport.left;
    miny = XBGI->viewport.top;
    maxx = XBGI->viewport.right;
    maxy = XBGI->viewport.bottom;
  }else{
    minx = 0;
    miny = 0;
    maxx = width-1;
    maxy = height-1;
  }
  p = 1;
  s[p][0] = Xx(x);
  s[p][1] = Xy(y);
  while ( p > 0 ){
    cx = s[p][0];
    cy = s[p--][1];
    if ( getpixel( xX(cx), yX(cy) ) != border ){
      while ( getpixel( xX(cx), yX(cy) )!=border && cx>=minx ) cx--;
      cx++;
      u = 0;
      d = 0;
      while ( getpixel( xX(cx), yX(cy) )!=border && cx<=maxx ){
        putpixel( xX(cx), yX(cy), XBGI->fillset.color );
        u2 = u;
        if ( cy > miny ){
          pu = getpixel( xX(cx), yX(cy-1) );
          if ( pu == border || pu == XBGI->fillset.color )
            u = 0;
          else
            u = 1;
        }
        if ( u2==0 && u==1 ){
          p++;
          s[p][0] = cx;
          s[p][1] = cy - 1;
        }
        d2 = d;
        if ( cy < maxy ) {
          pd = getpixel( xX(cx), yX(cy+1) );
          if ( pd == border || pd == XBGI->fillset.color )
            d = 0;
          else
            d = 1;
        }
        if ( d2==0 && d==1 ){
          p++;
          s[p][0] = cx;
          s[p][1] = cy + 1;
        }
        cx++;
      }
    }
  }
#else /* FLOODFILL_FAST_VERSION */
#define pix(x,y) mem[height*(y)+(x)]
  int s[1000][2];
  int p, cx, cy, u, u2, d, d2, pu, pd;
  int minx, miny, maxx, maxy;
  int width, height, i, j, k;
  XImage *image;
  unsigned char *data, *mem;

  getgeometry( &cx, &cy, &width, &height );
  mem = (unsigned char*)malloc( width*height*sizeof(char) );
  if ( mem == NULL ){
    XBGI->gerror = grNoFloodMem;
    return;
  }
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    0, 0, width, height, AllPlanes, ZPixmap );
  data = (unsigned char *)image->data;
  for ( i=0; i<height; i++ ){
    for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
      if ( j<image->xoffset ) continue;
      k++;
      if ( width < k ) continue;
      pix( k, i ) = XBGI->colortable[*data];
    }
  }
  XDestroyImage( image );
  /* start scan filling */
  if ( XBGI->viewport.clip ){
    minx = XBGI->viewport.left;
    miny = XBGI->viewport.top;
    maxx = XBGI->viewport.right;
    maxy = XBGI->viewport.bottom;
  }else{
    minx = 0;
    miny = 0;
    maxx = width-1;
    maxy = height-1;
  }
  p = 1;
  s[p][0] = Xx(x);
  s[p][1] = Xy(y);
  while ( p > 0 ){
    cx = s[p][0];
    cy = s[p--][1];
    if ( pix( cx, cy ) != border ){
      while ( pix( cx, cy )!=border && cx>=minx ) cx--;
      cx++;
      u = 0;
      d = 0;
      while ( pix( cx, cy )!=border && cx<=maxx ){
        putpixel( xX(cx), yX(cy), XBGI->fillset.color );
        u2 = u;
        if ( cy > miny ){
          pu = pix( cx, cy-1 );
          if ( pu == border || pu == XBGI->fillset.color )
            u = 0;
          else
            u = 1;
        }
        if ( u2==0 && u==1 ){
          p++;
          s[p][0] = cx;
          s[p][1] = cy - 1;
        }
        d2 = d;
        if ( cy < maxy ) {
          pd = pix( cx, cy+1 );
          if ( pd == border || pd == XBGI->fillset.color )
            d = 0;
          else
            d = 1;
        }
        if ( d2==0 && d==1 ){
          p++;
          s[p][0] = cx;
          s[p][1] = cy + 1;
        }
        cx++;
      }
    }
  }
  free( mem );
#undef pix
#endif
}

/*
 * functions for draw/fill rectangle
 */
void rectangle( int left, int top, int right, int bottom )
{
  XDrawRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left), abs(bottom-top) );
}
void bar( int left, int top, int right, int bottom )
{
  XFillRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left)+1, abs(bottom-top)+1 );
  /*XDrawRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                   min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                   abs(right-left), abs(bottom-top) );*/
}
void bar3d( int left, int top, int right, int bottom, int depth, int topflag )
{
  int topheight = 0;
  XPoint xpoints[4];

  if ( topflag ) topheight = depth/2;
  xpoints[0].x = Xx( right );         xpoints[0].y = Xy( top );
  xpoints[1].x = Xx( right + depth ); xpoints[1].y = Xy( top - topheight );
  xpoints[2].x = Xx( left + depth );  xpoints[2].y = Xy( top - topheight );
  xpoints[3].x = Xx( left );          xpoints[3].y = Xy( top );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
             xpoints, 4, CoordModeOrigin );
  xpoints[0].x = Xx( right );         xpoints[0].y = Xy( top );
  xpoints[1].x = Xx( right );         xpoints[1].y = Xy( bottom );
  xpoints[2].x = Xx( right + depth ); xpoints[2].y = Xy( bottom - topheight );
  xpoints[3].x = Xx( right + depth ); xpoints[3].y = Xy( top - topheight );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
             xpoints, 4, CoordModeOrigin );
  XFillRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left)+1, abs(bottom-top)+1 );
  XDrawRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left), abs(bottom-top) );
}

/*
 * functions for draw/fill poly
 */
static void polypoints2xpoints( int n, int polypoints[], XPoint *xpoints )
{
  int i;

  for ( i=0; i<n; i++ ){
    xpoints->x = Xx( *polypoints++ );
    xpoints->y = Xy( *polypoints++ );
    xpoints++;
  }
}
void drawpoly( int numpoints, int *polypoints )
{
  XPoint *xpoints = (XPoint *)malloc( numpoints * sizeof(XPoint) );
  
  polypoints2xpoints( numpoints, polypoints, xpoints );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
             xpoints, numpoints, CoordModeOrigin );
  free( xpoints );
}
void fillpoly( int numpoints, int *polypoints )
{
  XPoint *xpoints = (XPoint *)malloc( numpoints * sizeof(XPoint) );
  
  polypoints2xpoints( numpoints, polypoints, xpoints );
  XFillPolygon( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
               xpoints, numpoints, Complex, CoordModeOrigin );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
             xpoints, numpoints, CoordModeOrigin );
  free( xpoints );
}

/*
 * functions for draw/fill arc
 */
struct arcsettingstype {
  int x, y;
  int stangle, endangle;
  int xradius, yradius;
};
static struct arcsettingstype arcset = { 0, 0, 0, 360, 100, 100 };

void getaspectratio( int *xasp, int *yasp )
{
  *xasp = 10000; *yasp = 10000;
}
void setaspectratio( int xasp, int yasp )
{
  ReservedFuture( "setaspectratio()" );
}
void getarccoords( struct arccoordstype *arccoords )
{
  double stangle, endangle;
  
  stangle  = (double)arcset.stangle* M_PI/180;
  endangle = (double)arcset.endangle*M_PI/180;
  arccoords->x      = arcset.x;
  arccoords->y      = arcset.y;
  arccoords->xstart = arcset.x + (int)((double)arcset.xradius*cos(stangle ));
  arccoords->ystart = arcset.y - (int)((double)arcset.yradius*sin(stangle ));
  arccoords->xend   = arcset.x + (int)((double)arcset.xradius*cos(endangle));
  arccoords->yend   = arcset.y - (int)((double)arcset.yradius*sin(endangle));
}
void arc( int x, int y, int stangle, int endangle, int radius )
{
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = radius;
  arcset.yradius  = radius;
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           64 * stangle, 64 * ( endangle - stangle ) );
}
void pieslice( int x, int y, int stangle, int endangle, int radius )
{
  static struct arccoordstype arccoords;
  
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = radius;
  arcset.yradius  = radius;
  getarccoords( &arccoords );
  XFillArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           64 * stangle, 64 * ( endangle - stangle ) );
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           64 * stangle, 64 * ( endangle - stangle ) );
  line( arccoords.xstart, arccoords.ystart, arccoords.x, arccoords.y );
  line( arccoords.xend,   arccoords.yend,   arccoords.x, arccoords.y );
}
void circle( int x, int y, int radius )
{
#if 0
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = 0;
  arcset.endangle = 360;
  arcset.xradius  = radius;
  arcset.yradius  = radius;
#endif
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           0, 64 * ( 360 - 0 ) );
}
void ellipse( int x, int y, int stangle, int endangle,
             int xradius, int yradius )
{
#if 0
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = xradius;
  arcset.yradius  = yradius;
#endif
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - xradius ), Xy( y - yradius ), 2 * xradius, 2 * yradius,
           64 * stangle, 64 * ( endangle - stangle ) );
}
void sector( int x, int y, int stangle, int endangle,
            int xradius, int yradius )
{
#if 0
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = xradius;
  arcset.yradius  = yradius;
#endif
  XFillArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
           Xx( x - xradius ), Xy( y - yradius ), 2 * xradius, 2 * yradius,
           64 * stangle, 64 * ( endangle - stangle ) );
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - xradius ), Xy( y - yradius ), 2 * xradius, 2 * yradius,
           64 * stangle, 64 * ( endangle - stangle ) );
}
void fillellipse( int x, int y, int xradius, int yradius )
{
#if 0
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = 0;
  arcset.endangle = 360;
  arcset.xradius  = xradius;
  arcset.yradius  = yradius;
#endif
  XFillArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
           Xx( x - xradius ), Xy( y - yradius ) , 2 * xradius, 2 * yradius,
           0, 64 * ( 360 - 0 ) );
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - xradius ), Xy( y - yradius ) , 2 * xradius, 2 * yradius,
           0, 64 * ( 360 - 0 ) );
}

/*
 * functions for clear graphics
 */
void cleardevice( void )
{
  XClearWindow( XBGI->dpy, XBGI->win[XBGI->apage] );
}
void clearviewport( void )
{
  if ( XBGI->viewport.clip )
    XClearArea( XBGI->dpy, XBGI->win[XBGI->apage],
               XBGI->viewport.left, XBGI->viewport.top, 
               XBGI->viewport.right - XBGI->viewport.left + 1,
               XBGI->viewport.bottom - XBGI->viewport.top + 1, False );
  else
    cleardevice();
}

/*
 * functions for standard creation and termination graphics
 */
void detectgraph( int *graphdriver, int *graphmode )
{
  XVisualInfo vinfo_ret;
  Status result;
  
  /*
   * case of detecting before initgraph, always BGICOLOR
   */
  if ( XBGIS == NULL ){
    *graphdriver = BGI;
    *graphmode   = BGICOLOR;
    return;
  }
  /*
   * case of detecting after initgraph, research visual class and depth
   */
  switch( *graphdriver ){
  case DETECT:
    *graphdriver = XBGI->gdriver = BGI;
    *graphmode   = XBGI->gmode   = BGICOLOR;
    break;
  case BGI:
  default:
    *graphdriver = XBGI->gdriver = BGI;
    *graphmode   = XBGI->gmode   = BGICOLOR;
    break;
  }
  XBGI->depth  = DefaultDepth( XBGI->dpy, XBGI->scr );
  XBGI->vis    = DefaultVisual( XBGI->dpy, XBGI->scr );
  XBGI->vclass = XBGI->vis->class;
  switch ( XBGI->vclass ){
  case StaticGray:
  case GrayScale:
    *graphmode = XBGI->gmode = BGIGRAY;
    if ( XBGI->depth == 1 )
      *graphmode = XBGI->gmode = BGIMONO;
    break;
  case StaticColor:
  case TrueColor:
  case PseudoColor:
  case DirectColor:
    *graphmode = XBGI->gmode = BGICOLOR;
    break;
  }
}
void _initgraph( int argc, char *argv[], char *pd )
{
  _creategraph( argc, argv, pd );
}
void initgraph( int *graphdriver, int *graphmode, char *pathtodriver )
{
  creategraph( graphdriver, graphmode, pathtodriver );
}
int graphresult( void )
{
  return( XBGI->gerror );
}
void graphdefaults( void )
{
  int i;
  int x, y, width, height;

  for ( i=0; i<XBGI->npage; i++ ){
    setactivepage( i );
    getgeometry( &x, &y, &width, &height );
    setviewport( 0, 0, width-1, height-1, 0 ); 
    setcolor( XBGI->color );
    setbkcolor( XBGI->bkcolor );
    cleardevice();
  }
#ifdef 0
  setvisualpage( XBGI->vpage );
  setactivepage( XBGI->apage );
#else
  setvisualpage( 0 );
  setactivepage( 0 );
#endif
  setlinestyle( SOLID_LINE, 0, NORM_WIDTH );
  settextstyle( DEFAULT_FONT, HORIZ_DIR, 1 );
  settextjustify( LEFT_TEXT, TOP_TEXT );
  setfillstyle( SOLID_FILL, WHITE );
  moveto( 0, 0 );
}
char* grapherrormsg( int errorcode )
{
  ReservedFuture( "grapherrormsg()" );
  return( NULL );
}
void closegraph( void )
{
  destroygraph( XBGI );
}

/*
 * functions for set write function
 */
static int OP2GX[] = {
  GXcopy, GXxor, GXor, GXand, GXcopyInverted,
};
void setwritemode( int op )
{
  XSetFunction( XBGI->dpy, XBGI->gcdraw,  OP2GX[op] );
  XSetFunction( XBGI->dpy, XBGI->gcpixel, OP2GX[op] );
}

/*
 * functions for pixel
 */
unsigned getpixel( int x, int y )
{
  XImage *image;
  int val;

  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    Xx( x ), Xy( y ), 1, 1, AllPlanes, ZPixmap );
  val = XBGI->colortable[image->data[image->xoffset]];
  XDestroyImage( image );
  return( ( unsigned )val );
}
void putpixel( int x, int y, int color )
{
  XSetForeground( XBGI->dpy, XBGI->gcpixel,
                 XBGI->pixels[minmax(0,color,XBGI->palette.size-1)] );
  XDrawPoint( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcpixel, Xx(x), Xy(y) );
}

/*
 * functions for image
 */
unsigned imagesize( int left, int top, int right, int bottom )
{
  return( (unsigned)(2 + (abs(right-left)+1)*(abs(bottom-top)+1)) );
}
void getimage( int left, int top, int right, int bottom, void *bitmap )
{
  XImage *image;
  int width, height, i, j, k;
  unsigned char *data, *buffer = (unsigned char *)bitmap;

  width  = abs(right-left)+1;
  height = abs(bottom-top)+1;
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                    width, height, AllPlanes, ZPixmap );
  data = (unsigned char *)image->data;
  *buffer++ = width;
  *buffer++ = height;
  for ( i=0; i<height; i++ ){
    for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
      if ( j<image->xoffset ) continue;
      k++;
      if ( width < k ) continue;
      *buffer++ = XBGI->colortable[*data];
    }
  }
  XDestroyImage( image );
}
void putimage( int left, int top, void *bitmap, int op )
{
  int i;
  int width, height;
  XImage *image;
  unsigned char *src, *dst, *buffer = (unsigned char *)bitmap;
  
  width  = buffer[0];
  height = buffer[1];
  src = buffer + 2;
  dst = (unsigned char *)malloc( width*height*sizeof(unsigned char) );
  for ( i=0; i<width*height; i++ )
    dst[i] = XBGI->pixels[src[i]];
  XSetFunction( XBGI->dpy, XBGI->gcimage, OP2GX[minmax(COPY_PUT,op,NOT_PUT)] );
  image = XCreateImage( XBGI->dpy, DefaultVisual(XBGI->dpy,XBGI->scr),
                       XBGI->depth, ZPixmap, 0, (char *)dst,
                       width, height, 8, 0 );
  XPutImage( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcimage, image,
            0, 0, Xx(left), Xy(top), width, height );
  XDestroyImage( image );
  free( dst );
}

/*
 * functions for page
 */
int getmaxx( void )
{
  Window root;
  int x, y, width, height, border, depth;
  
  if ( XBGI->apage == 0 ){
    XGetGeometry( XBGI->dpy, XBGI->root, &root,
                 &x, &y, &width, &height, &border, &depth );
  }else{
    XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
                 &x, &y, &width, &height, &border, &depth );
  }
  return( width-1 );
}
int getmaxy( void )
{
  Window root;
  int x, y, width, height, border, depth;
  
  if ( XBGI->apage == 0 )
    XGetGeometry( XBGI->dpy, XBGI->root, &root,
                 &x, &y, &width, &height, &border, &depth );
  else
    XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
                 &x, &y, &width, &height, &border, &depth );
  return( height-1 );
}
void setactivepage( int page )
{
  XBGI->apage = minmax( 0, page, XBGI->npage-1 );
}
void setvisualpage( int page )
{
  int i;

  if ( page >= XBGI->npage ){
    /* all pages full mapped gmode */
    for ( i=0; i<XBGI->npage; i++ ){
      XReparentWindow( XBGI->dpy,
                      XBGI->win[i], RootWindow(XBGI->dpy,XBGI->scr), 0, 0 );
      XMapWindow( XBGI->dpy, XBGI->win[i] );
    }
    XUnmapWindow( XBGI->dpy, XBGI->root );
    XBGI->vpage = XBGI->npage;
  }else if ( ( XBGI->vpage >= XBGI->npage )&&( page < XBGI->npage ) ){
    /* one page mapped gmode */
    XUnmapSubwindows( XBGI->dpy, XBGI->root );
    for ( i=0; i<XBGI->npage; i++ )
      XReparentWindow( XBGI->dpy, XBGI->win[i], XBGI->root, 0, 0 );
    XBGI->vpage = minmax( 0, page, XBGI->npage-1 );
    XMapWindow( XBGI->dpy, XBGI->root );
    XMapRaised( XBGI->dpy, XBGI->win[XBGI->vpage] );
  }else{
    XBGI->vpage = minmax( 0, page, XBGI->npage-1 );
    XMapRaised( XBGI->dpy, XBGI->win[XBGI->vpage] );
  }
}

/*
 * functions for viewport
 */
void getviewsettings( struct viewporttype *viewport )
{
  *viewport = XBGI->viewport;
}
void setviewport( int left, int top, int right, int bottom, int clip )
{
  XRectangle rec;

  XBGI->viewport.left   = min( left, right );
  XBGI->viewport.top    = min( top, bottom );
  XBGI->viewport.right  = max( left, right );
  XBGI->viewport.bottom = max( top, bottom );
  XBGI->viewport.clip   = clip;
  if ( XBGI->viewport.clip ){
    rec.x = XBGI->viewport.left;
    rec.y = XBGI->viewport.top;
    rec.width  = XBGI->viewport.right - XBGI->viewport.left + 1;
    rec.height = XBGI->viewport.bottom - XBGI->viewport.top + 1;
    XSetClipRectangles( XBGI->dpy, XBGI->gcdraw,  0, 0, &rec, 1, YXSorted );
    XSetClipRectangles( XBGI->dpy, XBGI->gcfill,  0, 0, &rec, 1, YXSorted );
    XSetClipRectangles( XBGI->dpy, XBGI->gcpixel, 0, 0, &rec, 1, YXSorted );
    XSetClipRectangles( XBGI->dpy, XBGI->gcimage, 0, 0, &rec, 1, YXSorted );
  }else{
    XSetClipMask( XBGI->dpy, XBGI->gcdraw,  None );
    XSetClipMask( XBGI->dpy, XBGI->gcfill,  None );
    XSetClipMask( XBGI->dpy, XBGI->gcpixel, None );
    XSetClipMask( XBGI->dpy, XBGI->gcimage, None );
  }
}

/*
 * function for graph driver
 */
char* getdriversname( int driver_number )
{
  static char *drivers_name[BGI+1] = {
    "DETECT",
    "BGI",
  };

  switch( driver_number ){
  case DETECT:
  case BGI:
    return( drivers_name[driver_number] );
    break;
  }
  return( NULL );
}
char* getdrivername( void )
{
  return( getdriversname( XBGI->gdriver ) );
}
int installuserdriver( char *name, int (*detect)( void ) )
{
  ReservedFuture( "installuserdriver()" );
  return( 0 );
}

/*
 * functions for graph mode
 */
int getmaxmode( void )
{
  return( BGICOLOR );
}
int getgraphmode( void )
{
  return( XBGI->gmode );
}
char* getmodename( int mode_number )
{
  static char *mode_name[BGICOLOR+1] = {
    "BGIMONO",
    "BGIGRAY",
    "BGICOLOR",
  };

  switch( mode_number ){
  case BGIMONO:
  case BGIGRAY:
  case BGICOLOR:
    return( mode_name[mode_number] );
    break;
  }
  return( NULL );
}
void getmoderange( int graphdriver, int *lomode, int *himode )
{
  NotSupported( "getmoderange()" );
}
void setgraphmode( int gmode )
{
  XBGI->gmode = gmode;
  switch( XBGI->gmode ){
  case BGIMONO:
    setmaxcolor( 2 );
    break;
  case BGIGRAY:
    setmaxcolor( XBGI->palette.size );
    break;
  case BGICOLOR:
    setmaxcolor( XBGI->palette.size );
    break;
  case NOWINDOWMODE:
    ReservedFuture( "setgraphmode( NOWINDOWMODE )" );
    setmaxcolor( XBGI->palette.size );
    break;
  }
}
void restorecrtmode( void )
{
  NotSupported( "restorecrtmode()" );
}

/*
 * functions for color/palette
 */
int getmaxcolor( void )
{
  return( XBGI->palette.size-1 );
}
int getcolor( void )
{
  return( XBGI->color );
}
int getbkcolor( void )
{
  return( XBGI->bkcolor );
}
void setcolor( int color )
{
  XBGI->color = minmax( 0, color, XBGI->palette.size-1 );
  XSetForeground( XBGI->dpy, XBGI->gcdraw, XBGI->pixels[XBGI->color] );
}
void setbkcolor( int color )
{
  XBGI->bkcolor = minmax( 0, color, XBGI->palette.size-1 );
  XSetWindowBackground( XBGI->dpy, XBGI->win[XBGI->apage], 
                       XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcdraw,  XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcfill,  XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcpixel, XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcimage, XBGI->pixels[XBGI->bkcolor] );
  /*XClearWindow( XBGI->dpy, XBGI->win[XBGI->apage] );*/
}
int getpalettesize( void )
{
  return( XBGI->palette.size );
}
struct palettetype *getdefaultpalette( void )
{
  static struct palettetype def;
  int c;
  
  def.size = XBGI->palette.size;
  for ( c=0; c<def.size; c++ )
    def.colors[c] = c;
  return( &def );
}
void getpalette( struct palettetype *palette )
{
  *palette = XBGI->palette;
}
void setallpalette( struct palettetype *palette )
{
  int c;

  XBGI->palette = *palette;
  for ( c=0; c<palette->size; c++ )
    setpalette( palette->colors[c], c ); 
}
void _setrgbpalette( int colornum,
                    unsigned short red,
                    unsigned short green,
                    unsigned short blue )
{
  XColor xcolor;

  switch( XBGI->gmode ){
  case BGIMONO:
    break;
  case BGIGRAY:
  case BGICOLOR:
    colornum = minmax( 0, colornum, XBGI->palette.size-1 );
    XBGI->rgb[colornum].r = red;
    XBGI->rgb[colornum].g = green;
    XBGI->rgb[colornum].b = blue;
    xcolor.pixel = XBGI->pixels[colornum];
    xcolor.red   = XBGI->rgb[colornum].r;
    xcolor.green = XBGI->rgb[colornum].g;
    xcolor.blue  = XBGI->rgb[colornum].b;
    xcolor.flags = DoRed | DoGreen | DoBlue;
    XStoreColor( XBGI->dpy, XBGI->colormap, &xcolor );
    XFlush( XBGI->dpy );
    break;
  }
}
void setpalette( int colornum, int color )
{
  XColor xcolor;
  unsigned short red, green, blue;
  
  switch( XBGI->gmode ){
  case BGIMONO:
    break;
  case BGIGRAY:
  case BGICOLOR:
    colornum = minmax( 0, colornum, XBGI->palette.size-1 );
    color    = minmax( 0, color,    XBGI->palette.size-1 );
    XBGI->palette.colors[color] = (unsigned char)colornum;
    red   = XBGI->rgb[colornum].r;
    green = XBGI->rgb[colornum].g;
    blue  = XBGI->rgb[colornum].b;
    xcolor.pixel = XBGI->pixels[color];
    xcolor.red   = red;
    xcolor.green = green;
    xcolor.blue  = blue;
    xcolor.flags = DoRed | DoGreen | DoBlue;
    XStoreColor( XBGI->dpy, XBGI->colormap, &xcolor );
    XFlush( XBGI->dpy );
    break;
  }
}
void setrgbpalette( int colornum, int red, int green, int blue )
{
  _setrgbpalette( colornum, 
                 minmax( 0, 65536*red  /XBGI->palette.size, 65536-1 ),
                 minmax( 0, 65536*green/XBGI->palette.size, 65536-1 ),
                 minmax( 0, 65536*blue /XBGI->palette.size, 65536-1 ) );
}

/*
 * functions for fonts
 */
static char *FONTNAME[MAX_CHAR_FONT+1][MAX_CHAR_SIZE+1] = {
#include "fonts/adef.chr"
#include "fonts/trip.chr"
#include "fonts/litt.chr"
#include "fonts/sans.chr"
#include "fonts/goth.chr"
};
#ifdef ENABLE_BILINGUAL
static char *BFONTNAME[2][MAX_CHAR_SIZE+1] = {
#include "fonts/bdef.chr"
#include "fonts/jdef.chr"
};
#endif
#ifdef ENABLE_MULTILINGUAL
static char *FONTSETNAME[MAX_CHAR_SIZE+1] = {
#include "fonts/mdef.chr"
};
#endif
#ifdef ENABLE_BILINGUAL
#include "jutil.h"
static void XDrawBiString( Display *d, Drawable w, GC gc,
                          XFontStruct *afs, XFontStruct *bfs,
                          int x, int y, char *string, int len )
{
  unsigned char buf[2048], *ptr;
  XChar2b xchr;
  int ocode, code, c;
  unsigned jisc;

  ocode = jcode( string );
  str2jis( string, buf );
  if ( afs != NULL )
    XSetFont( d, gc, afs->fid );
  c = 0;
  ptr = buf;
  code = JCT_ANK;
  while ( *ptr != '\0' && c <= len ){
    if ( isjise(ptr) ){
      code = JCT_ANK;
      ptr += 3;
      if ( ocode == JCT_JIS ) c += 3;
    }else if ( isjiss(ptr) ){
      code = JCT_JIS;
      ptr += 3;
      if ( ocode == JCT_JIS ) c += 3;
    }else{
      if ( code==JCT_ANK ){
	if ( afs != NULL ){
	  XSetFont( d, gc, afs->fid );
	  XDrawString( d, w, gc, x, y, ptr, 1 );
	  x += XTextWidth( afs, ptr, 1 );
	}
        ptr++;
        c++;
      }else{
        jisc = s2code(ptr);
        xchr.byte1 = codec1(jisc);
        xchr.byte2 = codec2(jisc);      
	if ( bfs != NULL ){
	  XSetFont( d, gc, bfs->fid );
	  XDrawString16( d, w, gc, x, y, &xchr, 1 );
	  x += XTextWidth16( bfs, &xchr, 1 );
	}
        ptr += 2;
        c += 2;
      }
    }
  }
}
static int XBiTextWidth( XFontStruct *afs, XFontStruct *bfs,
                        char *string, int len )
{
  unsigned char buf[2048], *ptr;
  XChar2b xchr;
  int ocode, code, c;
  unsigned jisc;
  int x = 0;

  ocode = jcode( string );
  str2jis( string, buf );
  c = 0;
  ptr = buf;
  code = JCT_ANK;
  while ( *ptr != '\0' && c <= len ){
    if ( isjise(ptr) ){
      code = JCT_ANK;
      ptr += 3;
      if ( ocode == JCT_JIS ) c += 3;
    }else if ( isjiss(ptr) ){
      code = JCT_JIS;
      ptr += 3;
      if ( ocode == JCT_JIS ) c += 3;
    }else{
      if ( code==JCT_ANK ){
	if ( afs != NULL ){
	  x += XTextWidth( afs, ptr, 1 );
	}
        ptr++;
        c++;
      }else{
        jisc = s2code(ptr);
        xchr.byte1 = codec1(jisc);
        xchr.byte2 = codec2(jisc);      
	if ( bfs != NULL ){
	  x += XTextWidth16( bfs, &xchr, 1 );
	}
        ptr += 2;
        c += 2;
      }
    }
  }
  return( x );
}
#endif
void gettextsettings( struct textsettingstype *texttypeinfo )
{
  *texttypeinfo = XBGI->textset;
}
int installuserfont( char *name )
{
  ReservedFuture( "installuserfont()" );
  return( 0 );
}
void outtext( char *textstring )
{
  outtextxy( XBGI->cp.x, XBGI->cp.y, textstring );
  moverel( textwidth(textstring), 0 );
}
void outtextxy( int x, int y, char *textstring )
{
  int font = XBGI->textset.font;
  int dir  = XBGI->textset.direction;
  int size = XBGI->textset.charsize;
  int width = textwidth(textstring), height = textheight(textstring);
  int textx, texty;
  int i, j, k;
  Pixmap pixmap;
  XImage *image;
  unsigned char *data;
  GC gc;

  if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
    if ( !XBGI->chr[font][size].flag ) return;
  }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
    if ( !XBGI->chr[font][size].flag ) return;
#endif
#ifdef ENABLE_BILINGUAL
    /*if ( !XBGI->bchr0[0][size].flag || !XBGI->bchr0[1][size].flag ) return;*/
    /* font loading checking in XDrawBi */
#endif
#ifdef ENABLE_MULTILINGUAL
    if ( !XBGI->mchr0[size].flag ) return;
#endif
  }
  switch ( dir ){
  case VERT_DIR:
    /*
     * difficult vertical direction
     */
    pixmap = XCreatePixmap( XBGI->dpy, XBGI->win[XBGI->apage],
                           width, height, XBGI->depth );
    gc = XCreateGC( XBGI->dpy, pixmap, 0, 0 );
    XSetForeground( XBGI->dpy, gc, XBGI->pixels[XBGI->bkcolor] );
    XFillRectangle( XBGI->dpy, pixmap, gc, 0, 0, width, height );
    XSetForeground( XBGI->dpy, gc, XBGI->pixels[XBGI->color] );
    if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
      XSetFont( XBGI->dpy, gc, XBGI->chr[font][size].font->fid );
      XDrawString( XBGI->dpy, pixmap, gc,
                  0, XBGI->chr[font][size].font->max_bounds.ascent,
                  textstring, strlen(textstring) );
    }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
      XSetFont( XBGI->dpy, gc, XBGI->chr[font][size].font->fid );
      XDrawString( XBGI->dpy, pixmap, gc,
                  0, XBGI->chr[font][size].font->max_bounds.ascent,
                  textstring, strlen(textstring) );
#endif
#ifdef ENABLE_BILINGUAL
      XDrawBiString( XBGI->dpy, pixmap, gc,
                    XBGI->bchr0[0][size].font, XBGI->bchr0[1][size].font,
                    0,
                    max( XBGI->bchr0[0][size].font->max_bounds.ascent
                        ,XBGI->bchr0[1][size].font->max_bounds.ascent ),
                    textstring, strlen(textstring) );
#endif
#ifdef ENABLE_MULTILINGUAL
      XmbDrawString( XBGI->dpy, pixmap, XBGI->mchr0[size].font, gc, 0, 0,
                    textstring, strlen(textstring) );
#endif
    }
    if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
      switch( XBGI->textset.horiz ){
      case LEFT_TEXT:
        x += XBGI->chr[font][size].font->max_bounds.ascent;
        break;
      case CENTER_TEXT:
        x += height/2;
        break;
      case RIGHT_TEXT:
        x -= XBGI->chr[font][size].font->max_bounds.descent;
        break;
      }
    }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
      switch( XBGI->textset.horiz ){
      case LEFT_TEXT:
        x += XBGI->chr[font][size].font->max_bounds.ascent;
        break;
      case CENTER_TEXT:
        x += height/2;
        break;
      case RIGHT_TEXT:
        x -= XBGI->chr[font][size].font->max_bounds.descent;
        break;
      }
#endif
#ifdef ENABLE_BILINGUAL
      switch( XBGI->textset.horiz ){
      case LEFT_TEXT:
        x += max( XBGI->bchr0[0][size].font->max_bounds.ascent
                 ,XBGI->bchr0[1][size].font->max_bounds.ascent );
        break;
      case CENTER_TEXT:
        x += height/2;
        break;
      case RIGHT_TEXT:
        x -= max( XBGI->bchr0[0][size].font->max_bounds.descent
                 ,XBGI->bchr0[1][size].font->max_bounds.descent );
        break;
      }
#endif
#ifdef ENABLE_MULTILINGUAL
      switch( XBGI->textset.horiz ){
      case TOP_TEXT:
        break;
      case CENTER_TEXT:
        x -= textheight(textstring)/2;
        break;
      case BOTTOM_TEXT:
        x -= textheight(textstring);
        break;
      }
#endif
    }
    switch( XBGI->textset.vert ){
    case TOP_TEXT:
      break;
    case CENTER_TEXT:
      y -= width/2;
      break;
    case BOTTOM_TEXT:
      y -= width;
      break;
    }
    image = XGetImage( XBGI->dpy, pixmap, 0, 0, width, height,
                      AllPlanes, ZPixmap );
    data = (unsigned char *)image->data;
    for ( i=0; i<height; i++ ){
      for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
        if ( j<image->xoffset ) continue;
        k++;
        if ( width < k ) continue;
        if ( XBGI->colortable[*data] == XBGI->color )
          XDrawPoint( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                     x + i, y + ( width - k ) );
      }
    }
    XDestroyImage( image );
    XFreeGC( XBGI->dpy, gc );
    XFreePixmap( XBGI->dpy, pixmap );
    break;
  case HORIZ_DIR:
    /*
     * horizontal direction
     */
    switch( XBGI->textset.horiz ){
    case LEFT_TEXT:
      break;
    case CENTER_TEXT:
      x -= textwidth(textstring)/2;
      break;
    case RIGHT_TEXT:
      x -= textwidth(textstring);
      break;
    }
    if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
      switch( XBGI->textset.vert ){
      case TOP_TEXT:
        y += XBGI->chr[font][size].font->max_bounds.ascent;
        break;
      case CENTER_TEXT:
        y += height/2;
        break;
      case BOTTOM_TEXT:
        y -= XBGI->chr[font][size].font->max_bounds.descent;
        break;
      }
    }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
      switch( XBGI->textset.vert ){
      case TOP_TEXT:
        y += XBGI->chr[font][size].font->max_bounds.ascent;
        break;
      case CENTER_TEXT:
        y += height/2;
        break;
      case BOTTOM_TEXT:
        y -= XBGI->chr[font][size].font->max_bounds.descent;
        break;
      }
#endif
#ifdef ENABLE_BILINGUAL
      switch( XBGI->textset.vert ){
      case TOP_TEXT:
        y += max( XBGI->bchr0[0][size].font->max_bounds.ascent
                 ,XBGI->bchr0[1][size].font->max_bounds.ascent );
        break;
      case CENTER_TEXT:
        y += height/2;
        break;
      case BOTTOM_TEXT:
        y -= max( XBGI->bchr0[0][size].font->max_bounds.descent
                 ,XBGI->bchr0[1][size].font->max_bounds.descent );
        break;
      }
#endif
#ifdef ENABLE_MULTILINGUAL
      switch( XBGI->textset.vert ){
      case TOP_TEXT:
        y += textheight(textstring);
        break;
      case CENTER_TEXT:
        y -= textheight(textstring)/2;
        break;
      case BOTTOM_TEXT:
        y -= textheight(textstring);
        break;
      }
#endif
    }
    if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
      XDrawString( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                  Xx( x ), Xy( y ), textstring, strlen(textstring) );
    }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
      XDrawString( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                  Xx( x ), Xy( y ), textstring, strlen(textstring) );
#endif
#ifdef ENABLE_BILINGUAL
      XDrawBiString( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                    XBGI->bchr0[0][size].font, XBGI->bchr0[1][size].font,
                    Xx( x ), Xy( y ),
                    textstring, strlen(textstring) );
#endif
#ifdef ENABLE_MULTILINGUAL
      XmbDrawString( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->mchr0[size].font,
                    XBGI->gcdraw,
                    Xx( x ), Xy( y ), textstring, strlen(textstring) );
#endif
    }
    break;
  }
}
void settextjustify( int horiz, int vert )
{
  XBGI->textset.horiz = minmax( LEFT_TEXT, horiz, RIGHT_TEXT );
  XBGI->textset.vert  = minmax( BOTTOM_TEXT, vert, TOP_TEXT );
}
void settextstyle( int font, int direction, int charsize )
{
#ifdef ENABLE_MULTILINGUAL
  char **miss, *def;
  int n_miss;
#endif    

  font      = minmax( DEFAULT_FONT, font, MAX_CHAR_FONT );
  direction = minmax( HORIZ_DIR, direction, VERT_DIR );
  charsize  = minmax( USER_CHAR_SIZE, charsize, MAX_CHAR_SIZE );
  XBGI->textset.font      = font;
  XBGI->textset.direction = direction;
  XBGI->textset.charsize  = charsize;
  if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
    if ( !XBGI->chr[font][charsize].flag ){
      XBGI->chr[font][charsize].font
        = XLoadQueryFont( XBGI->dpy, FONTNAME[font][charsize] );
      if ( XBGI->chr[font][charsize].font != NULL )
	XBGI->chr[font][charsize].flag = !0;
      else
	XBGI->chr[font][charsize].flag = 0;
    }
    if ( XBGI->chr[font][charsize].font != NULL )
      XSetFont( XBGI->dpy, XBGI->gcdraw, XBGI->chr[font][charsize].font->fid );
  }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
    if ( !XBGI->chr[font][charsize].flag ){
      XBGI->chr[font][charsize].font
        = XLoadQueryFont( XBGI->dpy, FONTNAME[font][charsize] );
      if ( XBGI->chr[font][charsize].font != NULL )
	XBGI->chr[font][charsize].flag = !0;
      else
	XBGI->chr[font][charsize].flag = 0;
    }
    if ( XBGI->chr[font][charsize].font != NULL )
      XSetFont( XBGI->dpy, XBGI->gcdraw, XBGI->chr[font][charsize].font->fid );
#endif
#ifdef ENABLE_BILINGUAL
    if ( !XBGI->bchr0[0][charsize].flag || !XBGI->bchr0[1][charsize].flag ){
      XBGI->bchr0[0][charsize].font
        = XLoadQueryFont( XBGI->dpy, BFONTNAME[0][charsize] );
      XBGI->bchr0[1][charsize].font
        = XLoadQueryFont( XBGI->dpy, BFONTNAME[1][charsize] );
      if ( XBGI->bchr0[0][charsize].font != NULL )
	XBGI->bchr0[0][charsize].flag = !0;
      else
	XBGI->bchr0[0][charsize].flag = 0;
      if ( XBGI->bchr0[1][charsize].font != NULL )
	XBGI->bchr0[1][charsize].flag = !0;
      else
	XBGI->bchr0[1][charsize].flag = 0;
    }
#endif
#ifdef ENABLE_MULTILINGUAL
    if ( !XBGI->mchr0[charsize].flag ){
      XBGI->mchr0[charsize].font
        = XCreateFontSet( XBGI->dpy, FONTSETNAME[charsize],
                         &miss, &n_miss, &def );
      if ( n_miss == 0 )
	XBGI->mchr0[charsize].flag = !0;
      else
	XBGI->mchr0[charsize].flag = 0;
    }
#endif
  }
}
void setusercharsize( int multx, int divx, int multy, int divy )
{
  ReservedFuture( "setusercharsize()" );
}
int textheight( char *textstring )
{
  int font = XBGI->textset.font;
  int size = XBGI->textset.charsize;
#ifdef ENABLE_MULTILINGUAL
  XRectangle ink, log;
#endif

  if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
    if ( XBGI->chr[font][size].flag )
      return( XBGI->chr[font][size].font->max_bounds.ascent
	     +XBGI->chr[font][size].font->max_bounds.descent );
    else
      return 0;
  }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
    if ( XBGI->chr[font][size].flag )
      return( XBGI->chr[font][size].font->max_bounds.ascent
	     +XBGI->chr[font][size].font->max_bounds.descent );
    else
      return 0;
#endif
#ifdef ENABLE_BILINGUAL
    if ( XBGI->bchr0[0][size].flag && XBGI->bchr0[1][size].flag )
      return( max( XBGI->bchr0[0][size].font->max_bounds.ascent
		  +XBGI->bchr0[0][size].font->max_bounds.descent
		  ,XBGI->bchr0[1][size].font->max_bounds.ascent
		  +XBGI->bchr0[1][size].font->max_bounds.descent ) );
    else if ( XBGI->bchr0[0][size].flag )
      return( XBGI->bchr0[0][size].font->max_bounds.ascent
	     +XBGI->bchr0[0][size].font->max_bounds.descent );
    else if ( XBGI->bchr0[1][size].flag )
      return( XBGI->bchr0[1][size].font->max_bounds.ascent
	     +XBGI->bchr0[1][size].font->max_bounds.descent );
    else
      return 0;
#endif
#ifdef ENABLE_MULTILINGUAL
    if ( XBGI->mchr0[size].flag ){
      XmbTextExtents( XBGI->mchr0[size].font,
		     textstring, strlen(textstring), &ink, &log );
      return( ink.height );
    }else{
      return 0;
    }
#endif
  }
}
int textwidth( char *textstring )
{
  int font = XBGI->textset.font;
  int size = XBGI->textset.charsize;
  XRectangle ink, log;

  if ( XBGI->only_ank_flag || font != DEFAULT_FONT ){
    if ( XBGI->chr[font][size].flag )
      return( XTextWidth( XBGI->chr[font][size].font,
			 textstring, strlen(textstring)) );
    else
      return 0;
  }else{
#if !defined(ENABLE_BILINGUAL)&&!defined(ENABLE_MULTILINGUAL)
    if ( XBGI->chr[font][size].flag )
      return( XTextWidth( XBGI->chr[font][size].font,
			 textstring, strlen(textstring)) );
    else
      return 0;
#endif
#ifdef ENABLE_BILINGUAL
    /* if ( XBGI->bchr0[][size].flag ) is in XBiTextWidth */
    return( XBiTextWidth( XBGI->bchr0[0][size].font, XBGI->bchr0[1][size].font,
                         textstring, strlen(textstring)) );
#endif
#ifdef ENABLE_MULTILINGUAL
    if ( XBGI->mchr0[size].flag ){
      XmbTextExtents( XBGI->mchr0[size].font,
		     textstring, strlen(textstring), &ink, &log );
      return( ink.width );
    }else{
      return 0;
    }
#endif
  }
}

/*
 * internal functions to initialize for xbgi
 */
void XBGI_driver(void)
{
#include "bitmaps/icons/xbgi_a.xbm"
#include "bitmaps/icons/xbgi_b.xbm"
#include "bitmaps/icons/xbgi_c.xbm"
  struct xbm {
    char *bits;
    int width, height;
  } xbms[] = {
    { xbgi_c_bits, xbgi_c_width, xbgi_c_height },
    { xbgi_b_bits, xbgi_b_width, xbgi_b_height },
    { xbgi_a_bits, xbgi_a_width, xbgi_a_height },
    { NULL, 0, 0 }
  };
  int i, flag_mask, status;
  XSizeHints def;
  XSetWindowAttributes att;
  XEvent ev;
  XColor c1, c2;
  Pixmap bm;
  int argc = XBGI->argc;
  char **argv = XBGI->argv;

  char displayname[64] = "";
  char fnname[80]      = "lucidasans-bold-12";
  char fgname[20]      = "white";
  char bgname[20]      = "gray50";
  char defgeomstr[20]  = "640x400";
  char usrgeomstr[20]  = "640x400";
  char titlename[80]   = "";
  char iconname[80]    = "";
  char iconfile[80]    = "";
  
  XBGI->ncolor  = 16;
  XBGI->npage   = 2;
  XBGI->color   = XBGI->ncolor-1;
  XBGI->bkcolor = 0;
  XBGI->apage   = 0;
  XBGI->vpage   = 0;
  strcpy( titlename, argv[0] );
  strcpy( iconname,  argv[0] );
  XBGI->only_ank_flag = 0;

  /* get display name and initialization */
  scan_option( argc, argv, "-display %s", displayname );
  XBGI->dpy = XOpenDisplay( displayname );
  if ( XBGI->dpy == NULL ){
    XBGI->gerror = grNotDetected;
    ReservedFuture( "silent mode as without connection to display" );
    return;
  }
  XBGI->scr = DefaultScreen( XBGI->dpy );
  XBGI->colormap = DefaultColormap( XBGI->dpy, XBGI->scr );

  /* position, size, gc, font and color settings for work */
  scan_option( argc, argv, "-geometry %s", usrgeomstr );
  scan_option( argc, argv, "-fn %s", fnname );
  scan_option( argc, argv, "-fg %s", fgname );
  scan_option( argc, argv, "-bg %s", bgname );
  XBGI->gc = XCreateGC( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr), 0, NULL );
  XBGI->fs = XLoadQueryFont( XBGI->dpy, fnname );
  if ( XBGI->fs == NULL ){
    XBGI->gerror = grFontNotFound;
    ReservedFuture( "exception process as X cannot find font structure" );
  }else{
    XSetFont( XBGI->dpy, XBGI->gc, XBGI->fs->fid );
  }
  XAllocNamedColor( XBGI->dpy, XBGI->colormap, fgname, &c1, &c2 );
  XBGI->fg = c1.pixel;
  XAllocNamedColor( XBGI->dpy, XBGI->colormap, bgname, &c1, &c2 );
  XBGI->bg = c1.pixel;
  c2 = c1;
  c1.red   = min(  c1.red  *5/4,65535);
  c1.green = min(  c1.green*5/4,65535);
  c1.blue  = min(  c1.blue *5/4,65535);
  c2.red   = max(0,c2.red  *3/4      );
  c2.green = max(0,c2.green*3/4      );
  c2.blue  = max(0,c2.blue *3/4      );
  XAllocColor( XBGI->dpy, XBGI->colormap, &c1 );
  XAllocColor( XBGI->dpy, XBGI->colormap, &c2 );
  XBGI->lc = c1.pixel;
  XBGI->dc = c2.pixel;
  XSetForeground( XBGI->dpy, XBGI->gc, XBGI->fg );

  /* titlename and iconname settings */
  if ( scan_option( argc, argv, "-title %s", titlename ) )
    strcpy( iconname, titlename );
  scan_option( argc, argv, "-name %s", iconname );
  scan_option( argc, argv, "-icon %s", iconfile );

  /* get user bgi settings */
  scan_option( argc, argv, "-gd %d", &(XBGI->gdriver) );
  scan_option( argc, argv, "-gm %d", &(XBGI->gmode) );
  scan_option( argc, argv, "-ncolor  %d", &(XBGI->ncolor) );
  scan_option( argc, argv, "-npage %d", &(XBGI->npage) );
  XBGI->palette.size = (unsigned char)(XBGI->ncolor-1);
  scan_option( argc, argv, "-color %d",   &(XBGI->color)   );
  scan_option( argc, argv, "-bkcolor %d", &(XBGI->bkcolor) );
  scan_option( argc, argv, "-apage %d",   &(XBGI->apage)   );
  scan_option( argc, argv, "-vpage %d",   &(XBGI->vpage)   );

  if ( scan_option( argc, argv, "-only_ank" ) )  XBGI->only_ank_flag = !0;
  if ( scan_option( argc, argv, "-only_ank-" ) ) XBGI->only_ank_flag = 0;

#ifdef ENABLE_MULTILINGUAL
  if ( setlocal( LC_ALL, "" ) == NULL )
    XBGI->only_ank_flag = !0;
  else if ( XSupportsLocale() == False )
    XBGI->only_ank_flag = !0;
#endif

#ifdef ENABLE_ATEXIT
  if ( scan_option( argc, argv, "-atexit_loop" ) )
    atexit( xbgiloop );
#endif

  /* create windows */
  flag_mask = XGeometry( XBGI->dpy, XBGI->scr, usrgeomstr, defgeomstr,
                        0, 1, 1, 0, 0,
                        &def.x, &def.y, &def.width, &def.height );
  XBGI->root = XCreateSimpleWindow( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
                                   def.x, def.y, def.width, def.height, 0,
                                   WhitePixel(XBGI->dpy,XBGI->scr),
                                   BlackPixel(XBGI->dpy,XBGI->scr) );
  def.min_width  = 1;
  def.min_height = 1;
  def.max_width  = 2000;
  def.max_height = 2000;
  def.flags = PMinSize | PMaxSize;
  if ( flag_mask&XValue )      def.flags = def.flags | PPosition;
  if ( flag_mask&YValue )      def.flags = def.flags | USPosition;
  if ( flag_mask&WidthValue )  def.flags = def.flags | PSize;
  if ( flag_mask&HeightValue ) def.flags = def.flags | USSize;
  XSelectInput( XBGI->dpy, XBGI->root, ExposureMask|StructureNotifyMask );
  XSetStandardProperties( XBGI->dpy, XBGI->root,
                         titlename, iconname, None, argv, argc, &def );
  XSetWMNormalHints( XBGI->dpy, XBGI->root, &def );

  /* input event selection, GC creation and so on */
  XBGI->depth  = DefaultDepth( XBGI->dpy, XBGI->scr );
  XBGI->gcdraw  = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcfill  = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcpixel = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcimage = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  att.backing_store = Always;
  if ( scan_option( argc, argv, "-backing" ) ) 
    att.backing_store = Always;
  if ( scan_option( argc, argv, "-backing-" ) ) 
    att.backing_store = NotUseful;
  XBGI->win = (Window*)malloc( XBGI->npage*sizeof(Window) );
  for ( i=0; i<XBGI->npage; i++ ){
    XBGI->win[i] = XCreateSimpleWindow( XBGI->dpy, XBGI->root, 0, 0,
                                       def.width, def.height, 0,
                                       WhitePixel(XBGI->dpy,XBGI->scr),
                                       BlackPixel(XBGI->dpy,XBGI->scr) );
    XSelectInput( XBGI->dpy, XBGI->win[i], ExposureMask|
                 OwnerGrabButtonMask|ButtonPressMask|ButtonReleaseMask|
                 PointerMotionMask|ButtonMotionMask|KeyPressMask|
                 EnterWindowMask|LeaveWindowMask );
    XChangeWindowAttributes( XBGI->dpy, XBGI->win[i], CWBackingStore, &att );
    XSetStandardProperties( XBGI->dpy, XBGI->win[i],
                           titlename, iconname, None, argv, argc, &def );
  }
  
  /* set icon from data or file */
  if ( strcmp(iconfile,"")==0 ){
    for ( i=0; xbms[i].bits != NULL; i++ ){
      bm = XCreateBitmapFromData( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
                                 xbms[i].bits, xbms[i].width, xbms[i].height );
      status = XSetIconFromPixmap( XBGI->dpy, XBGI->root, bm,
                                  (unsigned int)xbms[i].width,
                                  (unsigned int)xbms[i].height );
      XFreePixmap( XBGI->dpy, bm );
      if ( status ) break;
    }
  }else{
    seticon( iconfile );
  }

  /* BGI default setting */
  detectgraph( &(XBGI->gdriver), &(XBGI->gmode) );
  setgraphmode( XBGI->gmode );
  graphdefaults();

  XMapWindow( XBGI->dpy, XBGI->root );
  XMapSubwindows( XBGI->dpy, XBGI->root ); 
  for ( i=0; i<XBGI->npage; i++ )
    do {
      XNextEvent( XBGI->dpy, &ev );
    } while ( Expose != ev.type );
  XFlush( XBGI->dpy );
  XBGI->gerror = grOk;
}

/*
 * other functions
 */
void _graphfreemem( void *ptr, unsigned size )
{
  NotSupported( "_graphfreemem()" );
}
void* _graphgetmem( unsigned size )
{
  NotSupported( "_graphgetmem()" );
  return( NULL );
}
unsigned setgraphbufsize( unsigned bufsize )
{
  NotSupported( "setgraphbufsize()" );
  return( 0 );
}

/*
 * save to x bitmap file
 */
int savegraphtoxbm( char *filename )
{
  int apage = XBGI->apage;
  Pixmap bitmap;
  GC gc;
  XImage *image;
  int i, j, k;
  int x, y, width, height;
  unsigned char *data;
  int status;

  setactivepage( XBGI->vpage );
  getgeometry( &x, &y, &width, &height );
  bitmap = XCreatePixmap( XBGI->dpy, XBGI->win[XBGI->apage],
                         width, height, 1 );
  gc = XCreateGC( XBGI->dpy, bitmap, 0, 0 );
  XSetForeground( XBGI->dpy, gc, WhitePixel(XBGI->dpy,XBGI->scr) );
  XSetBackground( XBGI->dpy, gc, BlackPixel(XBGI->dpy,XBGI->scr) );
  XSetFillStyle( XBGI->dpy, gc, FillSolid );
  XFillRectangle( XBGI->dpy, bitmap, gc, 0, 0, width, height );
  XSetForeground( XBGI->dpy, gc, BlackPixel(XBGI->dpy,XBGI->scr) );
  XSetBackground( XBGI->dpy, gc, WhitePixel(XBGI->dpy,XBGI->scr) );
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    0, 0, width, height, AllPlanes, ZPixmap );
  data = (unsigned char *)image->data;
  for ( i=0; i<height; i++ ){
    for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
      if ( j<image->xoffset ) continue;
      k++;
      if ( width < k ) continue;
      if ( XBGI->colortable[*data] != XBGI->bkcolor )
        XDrawPoint( XBGI->dpy, bitmap, gc, k, i );
    }
  }
  XDestroyImage( image );
  status = XWriteBitmapFile( XBGI->dpy, filename, bitmap,
                            width, height, 0, 0 );
  XFreeGC( XBGI->dpy, gc );
  XFreePixmap( XBGI->dpy, bitmap );
  setactivepage( apage );
  return status;
}
/*
 * save to x pixmap file
 */
int savegraphtoxpm( char *filename )
{
  XImage *image;
  int apage = XBGI->apage;
  int x, y, width, height;
  int status;

#ifdef ENABLE_XPMLIB
  setactivepage( XBGI->vpage );
  getgeometry( &x, &y, &width, &height );
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    0, 0, width, height, AllPlanes, ZPixmap );
  status = XpmWriteFileFromImage( XBGI->dpy, filename, image, 0, NULL );
  setactivepage( apage );
  XDestroyImage( image );
#else
  status = 0;
#endif
  return status;
}

/*
 * xbgiloop: standard loop with tools in XBGI
 */
/* menu defined */
enum MENU_ITEM {
  NO_SELECT = -1,
  SAVE_XBM,
  SAVE_XPM,
  QUIT,
};
static char *MENU_TITLE = "menu";
static int   MENU_ITEMC = 3;
static int   MENU_ITEMI = NO_SELECT;
#ifdef ENABLE_XPMLIB
static char *MENU_ITEMV[] = {
  "save bitmap",
  "save pixmap",
  "quit (q)"
};
#else
static char *MENU_ITEMV[] = {
  "save bitmap",
  "*save pixmap",
  "quit (q)"
  };
#endif
/* about and info. defined */
#define ABOUT "xbgi\nversion 2\ngraphics.h for X\nTaiji Yamada"
#define INFO  "id %d\ndriver %s\nmode %s\nvisual %d\nactive %d\npages %d\n\
color %d\nbkcolor %d\nncolor %d\n",\
XBGI->id, getdriversname(XBGI->gdriver), getmodename(XBGI->gmode),\
XBGI->apage, XBGI->vpage, XBGI->npage,\
XBGI->color, XBGI->bkcolor, XBGI->ncolor
/*
 * default event procs
 */
int XBGILOOP_button_press( int button, int x, int y )
{
  static char xbmfilename[80], xbmfilenamebak[80] = "scratch.xbm";
  static char xpmfilename[80], xpmfilenamebak[80] = "scratch.xpm";

  switch( button ){
  case 1:
    break;
  case 2:
    bgimenu( MENU_TITLE, MENU_ITEMC, MENU_ITEMV, &MENU_ITEMI );
    switch( MENU_ITEMI ){
    case NO_SELECT:
      break;
    case SAVE_XBM:
      strcpy( xbmfilename, xbmfilenamebak );
      bgigets( "XBM NAME", 80, xbmfilename );
      if ( strcmp(xbmfilename,"") != 0 ){
        savegraphtoxbm( xbmfilename );
        strcpy( xbmfilenamebak, xbmfilename );
      }else{
        strcpy( xbmfilename, xbmfilenamebak );
      }
      break;
    case SAVE_XPM:
      strcpy( xpmfilename, xpmfilenamebak );
      bgigets( "XPM NAME", 80, xpmfilename );
      if ( strcmp(xpmfilename,"") != 0 ){
        savegraphtoxpm( xpmfilename );
        strcpy( xpmfilenamebak, xpmfilename );
      }else{
        strcpy( xpmfilename, xpmfilenamebak );
      }
      break;
    case QUIT:
      return 0;
      break;
    }
    break;
  case 3:
    return 0;
    break;
  }
  return !0;
}
int XBGILOOP_button_release( int button, int x, int y )
{
  return !0;
}
int XBGILOOP_mouse_motion( int button, int x, int y )
{
  return !0;
}
int XBGILOOP_page_enter( int page )
{
  return !0;
}
int XBGILOOP_page_leave( int page )
{
  return !0;
}
int XBGILOOP_key_press( int c )
{
  int i;
  char buf[256];
  
  switch( c ){
  case CTRL('a'):
    bgiprintf( "about", CENTER_TEXT, ABOUT );
    break;
  case CTRL('i'):
    bgiprintf( "info.", LEFT_TEXT, INFO );
    break;
  case CTRL('o'):
    strcpy( buf, "" );
    for ( i=0; i<XBGI->argc; i++ ){
      if ( XBGI->argv[i][0] == '-' )
        strcat( buf, "\n" );
      strcat( buf, XBGI->argv[i] );
      if ( XBGI->argv[i][0] == '-' )
        strcat( buf, " " );
    }
    bgiprintf( "options", LEFT_TEXT, "%s", buf );
    break;
  case 'n': case 'N':
    setvisualpage( min( XBGI->vpage+1, XBGI->npage-1 ) );
    break;
  case 'b': case 'B':
    setvisualpage( XBGI->vpage-1 );
    break;
  case CTRL('f'):
    if ( XBGI->vpage != XBGI->npage )
      setvisualpage( XBGI->npage );
    else
      setvisualpage( 0 );
    break;
  case CTRL('c'):
    exit( 0 );
    break;
  case 'q': case 'Q':
    return 0;
    break;
  }
  return !0;
}
int XBGILOOP_configure( void )
{
  return !0;
}
int XBGILOOP_expose( void )
{
  return !0;
}
int XBGILOOP_noevent( void )
{
  return !0;
}

/*
 * event loop and tools
 */
void xbgiloop( void )
{
  xbgi *ptr;
  XEvent ev;
  int b, p;
  char c;
  
  /*
   * flush all xbgi
   */
  ptr = XBGIS;
  while ( ptr != NULL ){
    XFlush( ptr->dpy );
    ptr = ptr->next;
  }
  ptr = XBGI;
  XBGI = XBGIS;
  /*
   * finite event roop
   */
  while ( !0 ){
    if ( XBGI == NULL ){
      XBGI = XBGIS;
      continue;
    }
    if ( XEventsQueued( XBGI->dpy, QueuedAfterFlush ) > 0 ){
      XNextEvent( XBGI->dpy, &ev );
      switch ( ev.type ){
      case ButtonPress:
        XBGI->button_press_ev = ev;
        if ( XBGI->button_press_proc != NULL ){
          switch( ev.xbutton.button ){
          case Button1: b = 1; break;
          case Button2: b = 2; break;
          case Button3: b = 3; break;
          }
          if ( !XBGI->button_press_proc( b, ev.xbutton.x, ev.xbutton.y ) ){
            XBGI = ptr;
            return;
          }
        }
        break;
      case ButtonRelease:
        XBGI->button_release_ev = ev;
        if ( XBGI->button_release_proc != NULL ){
          switch( ev.xbutton.button ){
          case Button1: b = 1; break;
          case Button2: b = 2; break;
          case Button3: b = 3; break;
          }
          if ( !XBGI->button_release_proc( b, ev.xbutton.x, ev.xbutton.y ) ){
            XBGI = ptr;
            return;
          }
        }
        break;
      case MotionNotify:
        XBGI->mouse_motion_ev = ev;
        if ( XBGI->mouse_motion_proc != NULL ){
          b = 0;
          if ( ev.xmotion.state & Button1Mask ) b = 1;
          if ( ev.xmotion.state & Button2Mask ) b = 2;
          if ( ev.xmotion.state & Button3Mask ) b = 3;
          if ( !XBGI->mouse_motion_proc( b, ev.xbutton.x, ev.xbutton.y ) ){
            XBGI = ptr;
            return;
          }
        }
        break;
      case EnterNotify:
        XBGI->page_enter_ev = ev;
        if ( XBGI->page_enter_proc != NULL ){
          for ( p=0; p<XBGI->npage; p++ )
            if ( XBGI->win[p] == ev.xcrossing.window )
              if ( !XBGI->page_enter_proc( p ) ){
                XBGI = ptr;
                return;
              }
        }
        break;
      case LeaveNotify:
        XBGI->page_leave_ev = ev;
        if ( XBGI->page_leave_proc != NULL ){
          for ( p=0; p<XBGI->npage; p++ )
            if ( XBGI->win[p] == ev.xcrossing.window )
              if ( !XBGI->page_leave_proc( p ) ){
                XBGI = ptr;
                return;
              }
        }
        break;
      case KeyPress:
        if ( XLookupString( (XKeyEvent*)&ev, &c, 1, NULL, NULL ) == 1 ){
          XBGI->key_press_ev = ev;
          if ( XBGI->key_press_proc != NULL ){
            if ( !XBGI->key_press_proc( c ) ){
              XBGI = ptr;
              return;
            }
          }
        }
        break;
      case ConfigureNotify:
        if ( XBGI->configure_proc != NULL ){
          if ( !XBGI->configure_proc() ){
            XBGI = ptr;
            return;
          }
        }
        break;
      case Expose:
        if ( XBGI->expose_proc != NULL ){
          if ( !XBGI->expose_proc() ){
            XBGI = ptr;
            return;
          }
        }
        break;
      }
    }else{
      if ( XBGI->noevent_proc != NULL ){
        if ( !XBGI->noevent_proc() ){
          XBGI = ptr;
          return;
        }
      }
    }
    XBGI = XBGI->next;
  }
}

/*
 * following functions are customizing of xbgi and xbgiloop
 */

/***** event handler *****/
void set_button_press_proc( int (*proc)( int b, int x, int y ) )
{
  XBGI->button_press_proc = proc;
}
void set_button_release_proc( int (*proc)( int b, int x, int y ) )
{
  XBGI->button_release_proc = proc;
}
void set_mouse_motion_proc( int (*proc)( int b, int x, int y ) )
{
  XBGI->mouse_motion_proc = proc;
}
void set_page_enter_proc( int (*proc)( int p ) )
{
  XBGI->page_enter_proc = proc;
}
void set_page_leave_proc( int (*proc)( int p ) )
{
  XBGI->page_leave_proc = proc;
}
void set_key_press_proc( int (*proc)( int c ) )
{
  XBGI->key_press_proc = proc;
}
void set_configure_proc( int (*proc)( void )  )
{
  XBGI->configure_proc = proc;
}
void set_expose_proc( int (*proc)( void )  )
{
  XBGI->expose_proc = proc;
}
void set_noevent_proc( int (*proc)( void )  )
{
  XBGI->noevent_proc = proc;
}

/***** graphics drivers *****/
int registerbgidriver( void (*driver)(void) )
{
  XBGI_DRIVER = driver;
  return grOk;
}
int registerfarbgidriver( void *__driver ){ return grOk; }
int XBGI_driver_far[] = { 999 };

/***** graphics fonts *****/
int registerbgifont( void (*__font)(void) ){ return grOk; }
int registerfarbgifont( void *__font ){ return grOk; }
void triplex_font(){}
void small_font(){}
void sansserif_font(){}
void gothic_font(){}
int triplex_font_far[]   = { 999 };
int small_font_far[]     = { 999 };
int sansserif_font_far[] = { 999 };
int gothic_font_far[]    = { 999 };
