/*
 * Source File < graphix.c >
 *
 * Enhanced Functions: Borland Graphics Interface for Xlib
 *
 * Copyright (c) 1992-94 Tokyo Denki University, Taiji Yamada
 */
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "graphix.h"

#define XBGIENVIRON "XBGISET"
#define XBGISETFILE ".xbgisets"
#define SPACES " \t\n"

extern xbgi *XBGIS;
extern xbgi *XBGI;
extern void (*XBGI_DRIVER)(void);

xbgi *_creategraph( int argc, char *argv[], char *pd )
{
  static int id = 0;
  xbgi *this;
  int i, j;
  char buf[256], *ptr;
  
  this = (xbgi*)malloc(sizeof(xbgi));
  /*
   * insert to linked list
   */
  this->next = XBGIS;
  XBGIS = this;
  /*
   * set to current XBGI
   */
  XBGI = this;
  /*
   * set to parameter options
   */
  XBGI->argc = argc;
  XBGI->argv = (char**)malloc( argc*sizeof(char*) );
  for ( i=0; i<argc; i++ ){
    XBGI->argv[i] = (char*)malloc( (strlen(argv[i])+1)*sizeof(char) );
    strcpy( XBGI->argv[i], argv[i] );
  }
  /*
   * set to parameter options of pd
   */
  strcpy( buf, pd );
  ptr = strtok( buf, SPACES );
  while ( ptr != NULL ){
    XBGI->argc++;
    ptr = strtok( NULL, SPACES );
  }
  XBGI->argv = (char**)realloc( XBGI->argv, XBGI->argc*sizeof(char*) );
  strcpy( buf, pd );
  ptr = strtok( buf, SPACES );
  while ( ptr != NULL ){
    XBGI->argv[i] = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
    strcpy( XBGI->argv[argc++], ptr );
    ptr = strtok( NULL, SPACES );
  }
  /*
   * need initializing
   */
  this->id = id++;
  for ( i=DEFAULT_FONT; i<=MAX_CHAR_FONT; i++ )
    for ( j=USER_CHAR_SIZE; j<=MAX_CHAR_SIZE; j++ )
      this->chr[i][j].flag = 0;
#ifdef ENABLE_BILINGUAL
  for ( j=USER_CHAR_SIZE; j<=MAX_CHAR_SIZE; j++ ){
    this->bchr0[0][j].flag = 0;
    this->bchr0[1][j].flag = 0;
  }
#endif
#ifdef ENABLE_MULTILINGUAL
  for ( j=USER_CHAR_SIZE; j<=MAX_CHAR_SIZE; j++ )
    this->mchr0[j].flag = 0;
#endif
  /*
   * defualt event procedures in xbgiloop
   */
  XBGI->button_press_proc   = XBGILOOP_button_press;
  XBGI->button_release_proc = XBGILOOP_button_release;
  XBGI->mouse_motion_proc   = XBGILOOP_mouse_motion;
  XBGI->page_enter_proc     = XBGILOOP_page_enter;
  XBGI->page_leave_proc     = XBGILOOP_page_leave;
  XBGI->key_press_proc      = XBGILOOP_key_press;
  XBGI->configure_proc      = XBGILOOP_configure;
  XBGI->expose_proc         = XBGILOOP_expose;
  XBGI->noevent_proc        = XBGILOOP_noevent;

  this->gerror = grNoInitGraph;
  
  XBGI_DRIVER();

  return( this );
}
xbgi *creategraph( int *graphdriver, int *graphmode, char *pathtodriver )
{
  int argc;
  char **argv;
  int i, j;
  char *ptr, options[256], buf[256], buf2[256], file[64], key[64];
  FILE *fp;
  xbgi *this;

  /*
   * parameter settings to argc, argv from xbgienviron
   */
  sprintf( options, "%s -gd %d -gm %d ", "xbgi", *graphdriver, *graphmode );
  if ( getenv( XBGIENVIRON ) != NULL ){
    strcat( options, getenv( XBGIENVIRON ) );
    strcat( options, " " );
  }
  strcat( options, pathtodriver );
  strcpy( buf, options );
  ptr = strtok( buf, SPACES );
  argc = 0;
  while ( ptr != NULL ){
    argc++;
    ptr = strtok( NULL, SPACES );
  }
  argv = (char**)malloc( argc*sizeof(char*) );
  i = 0; 
  strcpy( buf, options );
  ptr = strtok( buf, SPACES );
  while ( ptr != NULL ){
    argv[i] = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
    strcpy( argv[i++], ptr );
    ptr = strtok( NULL, SPACES );
  }
  /*
   * parameter settings to argc, argv from xbgisetfile
   */
  strcpy( file, "" );
  ptr = getenv( "HOME" );
  if ( ptr != NULL ){
    strcpy( file, ptr );
    strcat( file, "/" );
    strcat( file, XBGISETFILE );
  }
  if ( (fp=fopen(file,"r")) != NULL ){
    j = i = strspn( pathtodriver, " " );
    if ( strcmp(pathtodriver,"") == 0 || pathtodriver[i] == '-' ){
      strcpy( key, "" );
    }else{
      while( pathtodriver[j]!=' ' && pathtodriver[j]!='\0' ) j++;
      strncat( key, &pathtodriver[i], j-i+1 );
    }
    while ( fgets( buf, 256, fp ) != NULL ){
      while ( buf[strlen(buf)-2] == '\\' ){
        buf[strlen(buf)-2] = '\0';
        if ( fgets( buf2, 256, fp ) != NULL )
          strcat( buf, buf2 );
        else
          break;
      }
      i = strspn( buf, " " );
      if ( ( strcmp( key, "" )==0 && buf[i] == '-' )
          || ( strncmp(&buf[i],key,strlen(key)) == 0 ) ){
        i = argc;
        strcpy( options, buf );
        ptr = strtok( options, SPACES );
        while ( ptr != NULL ){
          argc++;
          ptr = strtok( NULL, SPACES );
        }
        argv = (char**)realloc( argv, argc*sizeof(char*) );
        strcpy( options, buf );
        ptr = strtok( options, SPACES );
        while ( ptr != NULL ){
          argv[i] = (char*)malloc( (strlen(ptr)+1)*sizeof(char) );
          strcpy( argv[i++], ptr );
          ptr = strtok( NULL, SPACES );
        }
        break;
      }
    }
    fclose( fp );
  }
  /*
   * create xbgi
   */
  this = _creategraph( argc, argv, "" );
  for ( i=0; i<argc; i++ )
    free( argv[i] );
  free ( argv );
  return ( this );
}
xbgi *thisgraph( void )
{
  return( XBGI );
}
void currentgraph( xbgi *this )
{
  XBGI = this;
}
void destroygraph( xbgi *this )
{
  xbgi *ptr;

  XUnmapSubwindows( this->dpy, this->root );
  XUnmapWindow( this->dpy, this->root );
  XFlush( this->dpy );
  /*
   * delete from linked list
   */
  if ( XBGIS == this ){
    XBGIS = this->next;
    free( this );
    return;
  }
  ptr = XBGIS;
  while ( ptr->next != NULL ){
    if ( ptr->next == this ){
      ptr->next = this->next;
      XBGI = ptr;
      free( this );
      return;
    }
    ptr = ptr->next;
  }
  return;
}
